import wgpu  # only for flags/enums

from ....utils import array_from_shadertype
from ....resources import Buffer, Texture
from ....objects import Points
from ....materials import (
    PointsMaterial,
    PointsGaussianBlobMaterial,
    PointsMarkerMaterial,
    PointsSpriteMaterial,
)
from ....utils.enums import MarkerInt, MarkerShape


from .. import (
    register_wgpu_render_function,
    BaseShader,
    Binding,
    load_wgsl,
    nchannels_from_format,
    to_texture_format,
    GfxSampler,
    GfxTextureView,
)

renderer_uniform_type = dict(last_i="i4")


@register_wgpu_render_function(Points, PointsMaterial)
class PointsShader(BaseShader):
    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)
        material = wobject.material
        geometry = wobject.geometry

        color_mode = str(material.color_mode).split(".")[-1]
        if color_mode == "auto":
            if material.map is not None:
                self["color_mode"] = "vertex_map"
                self["color_buffer_channels"] = 0
            else:
                self["color_mode"] = "uniform"
                self["color_buffer_channels"] = 0
        elif color_mode == "uniform":
            self["color_mode"] = "uniform"
            self["color_buffer_channels"] = 0
        elif color_mode == "vertex":
            nchannels = nchannels_from_format(geometry.colors.format)
            self["color_mode"] = "vertex"
            self["color_buffer_channels"] = nchannels
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "vertex_map":
            self["color_mode"] = "vertex_map"
            self["color_buffer_channels"] = 0
            if material.map is None:
                raise ValueError("Cannot apply colormap is no material.map is set.")
        elif color_mode == "debug":
            self["color_mode"] = "debug"
            self["color_buffer_channels"] = 0
        else:
            raise RuntimeError(f"Unknown color_mode: '{color_mode}'")

        if isinstance(material, PointsMarkerMaterial):
            edge_color_mode = str(material.edge_color_mode).split(".")[-1]
        else:
            edge_color_mode = "uniform"

        if edge_color_mode == "vertex":
            nchannels = nchannels_from_format(geometry.edge_colors.format)
            self["edge_color_mode"] = "vertex"
            self["edge_color_buffer_channels"] = nchannels
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(
                    f"Geometry.edge_colors needs 1-4 columns, not {self['edge_color_buffer_channels']}"
                )
        else:  # auto or uniform
            self["edge_color_mode"] = "uniform"
            self["edge_color_buffer_channels"] = 0

        self["edge_mode"] = material.edge_mode
        self["rotation_mode"] = material.rotation_mode
        self["is_sprite"] = 0  # 0, 1, 2
        self["need_neighbours"] = material.rotation_mode == "curve"

        if isinstance(material, PointsSpriteMaterial):
            self["is_sprite"] = 1
            if material.sprite is not None:
                self["is_sprite"] = 2  # i.e. is sprite and has a texture

        self["size_mode"] = str(material.size_mode).split(".")[-1]
        self["size_space"] = material.size_space
        self["aa"] = material._gfx_effective_aa

        self["marker_mode"] = ""
        self["draw_line_on_edge"] = False
        self["is_gaussian"] = False
        if isinstance(material, PointsGaussianBlobMaterial):
            self["is_gaussian"] = True
        elif isinstance(material, PointsMarkerMaterial):
            self["marker_mode"] = material.marker_mode
            self["uniform_marker"] = MarkerInt[material.marker]
            custom_sdf = material.custom_sdf
            if custom_sdf is None:
                # Make a nice full square to help the user better design their
                # custom SDF
                custom_sdf = "return max(abs(coord.x), abs(coord.y)) - size * 0.5;"
            self["custom_sdf"] = custom_sdf
            self["draw_line_on_edge"] = True
            for marker_name in MarkerShape:
                self[f"markerenum_{marker_name}"] = MarkerInt[marker_name]

    def get_bindings(self, wobject, shared):
        geometry = wobject.geometry
        material = wobject.material

        rbuffer = "buffer/read_only_storage"
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
            Binding("s_positions", rbuffer, geometry.positions, "VERTEX"),
        ]

        if self["need_neighbours"]:
            uniform_buffer = Buffer(
                array_from_shadertype(renderer_uniform_type), force_contiguous=True
            )
            uniform_buffer.data["last_i"] = geometry.positions.nitems - 1
            bindings.append(
                Binding("u_renderer", "buffer/uniform", uniform_buffer),
            )

        if self["size_mode"] == "vertex":
            bindings.append(Binding("s_sizes", rbuffer, geometry.sizes, "VERTEX"))

        # Per-vertex color, colormap, or a uniform color?
        if self["color_mode"] == "vertex":
            bindings.append(Binding("s_colors", rbuffer, geometry.colors, "VERTEX"))
        elif self["color_mode"] == "vertex_map":
            bindings.append(
                Binding("s_texcoords", rbuffer, geometry.texcoords, "VERTEX")
            )
            bindings.extend(
                self.define_generic_colormap(material.map, geometry.texcoords)
            )

        if self["edge_color_mode"] == "vertex":
            bindings.append(
                Binding("s_edge_colors", rbuffer, geometry.edge_colors, "VERTEX")
            )

        if self["rotation_mode"] == "vertex":
            bindings.append(
                Binding("s_rotations", rbuffer, geometry.rotations, "VERTEX")
            )

        if self["rotation_mode"] == "vertex":
            bindings.append(
                Binding("s_rotations", rbuffer, geometry.rotations, "VERTEX")
            )

        if self["marker_mode"] == "vertex":
            bindings.append(Binding("s_markers", rbuffer, geometry.markers, "VERTEX"))

        # Process sprite texture. Note that we can *also* have a colormap for the base color.
        if self["is_sprite"] == 2:
            sprite_sampler = GfxSampler("linear", "clamp")
            if not isinstance(material.sprite, Texture):
                raise TypeError("material sprite must be a Texture")
            sprite_view = GfxTextureView(material.sprite)
            if sprite_view.view_dim != "2d":
                raise ValueError("Sprite textures must be 2D")
            fmt = to_texture_format(sprite_view.format)
            if not ("norm" in fmt or "float" in fmt):
                raise ValueError("Sprite textures must be u8norm or float")
            self["sprite_nchannels"] = len(fmt) - len(fmt.lstrip("rgba"))
            bindings += [
                Binding("s_sprite", "sampler/filtering", sprite_sampler, "FRAGMENT"),
                Binding("t_sprite", "texture/auto", sprite_view, "FRAGMENT"),
            ]

        bindings = {i: b for i, b in enumerate(bindings)}
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_list,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        offset, size = wobject.geometry.positions.draw_range
        offset, size = offset * 6, size * 6
        return {
            "indices": (size, 1, offset, 0),
        }

    def get_code(self):
        return load_wgsl("points.wgsl")



import numpy as np

from ._base import WorldObject
from ..resources import Buffer
from ..utils import unpack_bitfield, array_from_shadertype
from ..materials import BackgroundMaterial


class Group(WorldObject):
    """A group of objects.

    A Group is useful when manipulating the scene graph as children can be
    jointly moved/scaled/rotated. It has no visual properties.

    Parameters
    ----------
    visible : bool
        If true, the object and its children are visible.

    name : str
        The name of the group.

    """

    def __init__(self, *, visible=True, name=""):
        super().__init__(visible=visible, name=name)


class Scene(Group):
    """Root of the scene graph.

    The scene holds scene-level information (background color, fog, environment
    map) as well as all objects that take part in the rendering process as
    either direct or indirect children/nested objects.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Background(WorldObject):
    """The scene's background.

    Can be e.g. a gradient, a static image or a skybox.

    Parameters
    ----------
    geometry : Geometry
        Must be ``None``. Exists for compliance with the generic WorldObject
        API.
    material : Material
        The material to use when rendering the background.
    kwargs : Any
        Additional kwargs are forwarded to the object's :class:`base class
        <pygfx.objects.WorldObject>`.

    """

    def __init__(self, geometry=None, material=None, *, render_order=-1e6, **kwargs):
        if geometry is not None and material is None:
            raise TypeError("You need to instantiate using Background(None, material)")
        super().__init__(None, material, render_order=render_order, **kwargs)

    @classmethod
    def from_color(cls, *colors):
        """Create a background with a :class:`.BackgroundMaterial`,
        using 1 uniform color, 2 colors for a vertical gradient, or 4 colors (one for each corner).
        """
        return cls(None, BackgroundMaterial(*colors))


class Grid(WorldObject):
    """A grid to help interpret spatial distances.

    The grid by default occupies 1x1 square in the xz plane. It can be scaled,
    rotated, and translated to move it into position. If the grid is infinite
    (``material.infinite``) then scale and in-plane translations are ignored.

    Parameters
    ----------
    geometry : Geometry
        Must be ``None``. Exists for compliance with the generic WorldObject API.
    material : Material
        The material to use when rendering the background.
    orientation : str
        The (initial) grid rotation. Must be 'xy' (default), 'xz', or 'yz'.
        Simply rotates the object, e.g. for 'xz' will do ``self.local.euler_x = np.pi/2``.
    kwargs : Any
        Additional kwargs are forwarded to the object's :class:`base class
        <pygfx.objects.WorldObject>`.
    """

    def __init__(
        self, geometry=None, material=None, *, orientation, render_order=-100, **kwargs
    ):
        if geometry is not None and material is None:
            raise TypeError("You need to instantiate using Grid(None, material)")
        super().__init__(None, material, render_order=render_order, **kwargs)
        if orientation is not None:
            if orientation == "xy":
                pass
            elif orientation == "xz":
                self.local.euler_x = np.pi / 2
            elif orientation == "yz":
                self.local.euler_y = -np.pi / 2
            else:
                raise ValueError(
                    f"Invalid grid orientation: '{orientation}', must be 'xz', 'xy', or 'yz'."
                )


class Line(WorldObject):
    """An object representing a line using a list of vertices (3D positions).

    Some materials will render the line as a continuous line, while other materials
    will consider each pair of consecutive points a segment.

    The picking info of a Line (the result of ``renderer.get_pick_info()``) will
    for most materials include ``vertex_index`` (int) and ``segment_coord``
    (float, sub-segment coordinate).

    Parameters
    ----------
    geometry : Geometry
        The data defining the shape of the object. Must contain at least a
        "positions" buffer. Depending on the usage of the material, can also
        include buffers "texcoords", "colors", and "sizes".
    material : Material
        The data defining the appearance of the object.
    visible : bool
        Whether the object is visible.
    render_order : int
        Affects the order in which objects are rendered.
    position : Vector
        The position of the object in the world. Default (0, 0, 0).

    """


class Points(WorldObject):
    """A point cloud.

    An object consisting of points represented by vertices (3D positions).

    The picking info of a Points object (the result of
    ``renderer.get_pick_info()``) will for most materials include
    ``vertex_index`` (int) and ``point_coord`` (tuple of 2 float coordinates in
    logical pixels).

    Parameters
    ----------
    geometry : Geometry
        The data defining the shape of the object. Must contain at least a
        "positions" buffer. Depending on the usage of the material, can also
        include buffers "texcoords", "colors", "sizes", "edge_colors", "rotations".
    material : Material
        The data defining the appearance of the object.
    visible : bool
        Whether the object is visible.
    render_order : int
        Affects the order in which objects are rendered.
    position : Vector
        The position of the object in the world. Default (0, 0, 0).

    """


class Mesh(WorldObject):
    """A mesh.

    An object consisting of triangular faces represented by a set of vertices
    (3D positions) and a set of vertex indices indicating which vertex triplets
    form mesh triangles.

    The picking info of a Mesh (the result of ``renderer.get_pick_info()``) will
    for most materials include ``instance_index`` (int), ``face_index`` (int),
    and ``face_coord`` (tuple of 3 floats). The latter are the barycentric
    coordinates for each vertex of the face (with values 0..1).

    Parameters
    ----------
    geometry : Geometry
        The data defining the shape of the object. Must contain at least a
        "positions" and "indices" buffers. Depending on the usage of the material, can also
        include buffers "normals", "colors", "texcoords", "texcoords2", etc.
        Advanced use of meshes also support "tangents", "skin_indices",  and "skin_weights".
    material : Material
        The data defining the appearance of the object.
    visible : bool
        Whether the object is visible.
    render_order : int
        Affects the order in which objects are rendered.
    position : Vector
        The position of the object in the world. Default (0, 0, 0).

    """

    def __init__(self, geometry=None, material=None, *args, **kwargs):
        super().__init__(geometry, material, *args, **kwargs)
        self._morph_target_influences = None
        self._morph_target_names = []

    @property
    def morph_target_influences(self):
        """
        An ndarray of weights typically from 0-1 that specify how much of the morph is applied.

        Note:
        When using this attribute, its geometry needs to have the relevant attributes of morph targets.
        """
        return self._morph_target_influences.data["influence"][:-1]

    @morph_target_influences.setter
    def morph_target_influences(self, value):
        value = np.asarray(value, dtype=np.float32)

        # Get the number of morph targets from the morph data on the geometry
        morph_attrs = [
            getattr(self.geometry, name, None)
            for name in ["morph_positions", "morph_normals", "morph_colors"]
        ]
        morph_attrs = [x for x in morph_attrs if x is not None]
        if not morph_attrs:
            return
        morph_count = min(len(x) for x in morph_attrs)

        # Check with the size of the given data
        if len(value) != morph_count:
            raise ValueError(
                f"Length of morph target influences must match the number of morph targets. Expected {morph_count}, got {len(value)}."
            )

        buffer_size = morph_count + 1  # add roon for base influence
        if (
            self._morph_target_influences is None
            or self._morph_target_influences.nitems != buffer_size
        ):
            self._morph_target_influences = Buffer(
                array_from_shadertype(
                    {
                        "influence": "f4",
                    },
                    buffer_size,
                )
            )

        if getattr(self.geometry, "morph_targets_relative", False):
            base_influence = 1.0
        else:
            base_influence = 1 - value.sum()

        self._morph_target_influences.data["influence"][:-1] = value
        self._morph_target_influences.data["influence"][-1] = base_influence
        self._morph_target_influences.update_full()

    @property
    def morph_target_names(self):
        """
        A list of names for the morph targets.
        """
        return self._morph_target_names

    def _wgpu_get_pick_info(self, pick_value) -> dict:
        info = super()._wgpu_get_pick_info(pick_value)
        values = unpack_bitfield(
            pick_value, wobject_id=20, index=26, coord1=9, coord2=9
        )
        face_index = values["index"]
        face_coord = [
            values["coord1"] / 511,
            values["coord2"] / 511,
        ]
        # The shader encodes only two of the barycentric coordinates. The third can be computed, as they add up to one.
        face_coord.append(1.0 - face_coord[0] - face_coord[1])
        if (
            self.geometry.indices.data is not None
            and self.geometry.indices.data.shape[-1] == 4
        ):
            triangle_index = face_index % 2
            face_index = face_index // 2
            if triangle_index == 0:
                # The sub indices are 0, 1, 2, so we just add the zero for index 3.
                face_coord.append(0.0)
            else:
                # The sub indices are 0, 2, 3. The index 3 uses the
                # face_coord slot of index 1, (see meshshader.py), so
                # we put that at the end and put a zero in its place.
                face_coord = face_coord[0], 0.0, face_coord[2], face_coord[1]
        info["face_index"] = face_index
        info["face_coord"] = tuple(face_coord)
        return info


class Image(WorldObject):
    """A 2D image.

    The geometry for this object consists only of ``geometry.grid``: a
    texture with the 2D data.

    If no colormap is applied to the material, the data are interpreted as
    colors in sRGB space. To use physical space instead, set the texture's
    colorspace property to ``"physical"``.

    The picking info of an Image (the result of ``renderer.get_pick_info()``)
    will for most materials include ``index`` (tuple of 2 int), and
    ``pixel_coord`` (tuple of float subpixel coordinates).

    Parameters
    ----------
    geometry : Geometry
        The data defining the shape of the object. Must contain at least a
        "grid" attribute for a 2D texture.
    material : Material
        The data defining the appearance of the object.
    visible : bool
        Whether the object is visible.
    render_order : int
        Affects the order in which objects are rendered.
    position : Vector
        The position of the object in the world. Default (0, 0, 0).

    """

    def _wgpu_get_pick_info(self, pick_value) -> dict:
        info = super()._wgpu_get_pick_info(pick_value)
        tex = self.geometry.grid
        if hasattr(tex, "texture"):
            tex = tex.texture  # tex was a view
        # This should match with the shader
        values = unpack_bitfield(pick_value, wobject_id=20, x=22, y=22)
        x = values["x"] / 4194303 * tex.size[0] - 0.5
        y = values["y"] / 4194303 * tex.size[1] - 0.5
        ix, iy = int(x + 0.5), int(y + 0.5)
        info["index"] = (ix, iy)
        info["pixel_coord"] = (x - ix, y - iy)
        return info


class Volume(WorldObject):
    """A 3D image.

    The geometry for this object consists only of ``geometry.grid``: a texture
    with the 3D data.

    The picking info of a Volume (the result of ``renderer.get_pick_info()``)
    will for most materials include ``index`` (tuple of 3 int), and
    ``voxel_coord`` (tuple of float subpixel coordinates).

    Parameters
    ----------
    geometry : Geometry
        The data defining the shape of the object. Must contain at least a
        "grid" attribute for a 3D texture.
    material : Material
        The data defining the appearance of the object.
    visible : bool
        Whether the object is visible.
    render_order : int
        Affects the order in which objects are rendered.
    position : Vector
        The position of the object in the world. Default (0, 0, 0).

    """

    def _wgpu_get_pick_info(self, pick_value) -> dict:
        info = super()._wgpu_get_pick_info(pick_value)
        tex = self.geometry.grid
        if hasattr(tex, "texture"):
            tex = tex.texture  # tex was a view
        # This should match with the shader
        values = unpack_bitfield(pick_value, wobject_id=20, x=14, y=14, z=14)
        texcoords_encoded = values["x"], values["y"], values["z"]
        size = tex.size
        x, y, z = [
            (v / 16383) * s - 0.5 for v, s in zip(texcoords_encoded, size, strict=True)
        ]
        ix, iy, iz = int(x + 0.5), int(y + 0.5), int(z + 0.5)
        info["index"] = (ix, iy, iz)
        info["voxel_coord"] = (x - ix, y - iy, z - iz)
        return info



from __future__ import annotations

from typing import TYPE_CHECKING

from ..utils.trackable import Trackable
from ..utils import array_from_shadertype, ReadOnlyDict
from ..resources import Buffer
from ..utils.enums import AlphaMethod, AlphaMode


if TYPE_CHECKING:
    from typing import Optional, ClassVar, TypeAlias, Literal, Sequence

    ABCDTuple: TypeAlias = tuple[float, float, float, float]


TEST_COMPARE_VALUES = "<", "<=", "==", "!=", ">=", ">"  # for alpha_test and depth_test


ALPHA_MODES = {
    "solid": {
        "method": "opaque",
        "premultiply_alpha": False,
    },
    "solid_premul": {
        "method": "opaque",
        "premultiply_alpha": True,
    },
    "blend": {
        "method": "blended",
        "color_op": "add",
        "alpha_op": "add",
        "color_constant": (0, 0, 0),
        "alpha_constant": 0,
        "color_src": "src-alpha",
        "color_dst": "one-minus-src-alpha",
        "alpha_src": "one",
        "alpha_dst": "one-minus-src-alpha",
    },
    "add": {
        "method": "blended",
        "color_op": "add",
        "alpha_op": "add",
        "color_constant": (0, 0, 0),
        "alpha_constant": 0,
        "color_src": "src-alpha",
        "color_dst": "one",
        "alpha_src": "src-alpha",
        "alpha_dst": "one",
    },
    "subtract": {
        "method": "blended",
        "color_op": "add",
        "alpha_op": "add",
        "color_constant": (0, 0, 0),
        "alpha_constant": 0,
        "color_src": "zero",
        "color_dst": "one-minus-src",
        "alpha_src": "zero",
        "alpha_dst": "one",
    },
    "multiply": {
        "method": "blended",
        "color_op": "add",
        "alpha_op": "add",
        "color_constant": (0, 0, 0),
        "alpha_constant": 0,
        "color_src": "zero",
        "color_dst": "src",
        "alpha_src": "zero",
        "alpha_dst": "src",
    },
    "weighted_blend": {
        "method": "weighted",
        "weight": "alpha",
        "alpha": "alpha",
    },
    "weighted_solid": {
        "method": "weighted",
        "weight": "alpha",
        "alpha": "1.0",
    },
    "dither": {
        "method": "stochastic",
        "pattern": "blue_noise",
        "seed": "element",
    },
    "bayer": {
        "method": "stochastic",
        "pattern": "bayer",
        "seed": "object",  # bc not enough variation to do 'element'
    },
}


class Material(Trackable):
    """Material base class.

    Parameters
    ----------
    opacity : float
        The opacity (a.k.a. alpha value) applied to this material, expressed as
        a value between 0 and 1. If the material contains any non-opaque
        fragments, their alphas are simply scaled by this value.
    clipping_planes : tuple
        A tuple of planes (abcd tuples) in world space. Points in space whose
        signed distance to the plane is negative are clipped (not rendered).
        Applies to the object to which this material is attached.
    clipping_mode : str
        Set the behavior for multiple clipping planes: "any" or "all". If this
        is "any" (the default) a fragment is discarded if it is clipped by any
        clipping plane. If this is "all", a fragment is discarded only if it is
        clipped by *all* of the clipping planes.
    alpha_mode : str
        How the alpha value of an object is used to to combine the resulting color
        with the target color texture.
    alpha_config : dict
        An advanced way to fully control the alpha behaviour. When both ``alpha_mode``
        and ``alpha_config`` are given, the latter is used.
    depth_test :  bool
        Whether the object takes the depth buffer into account (and how).
        Default True.
    depth_compare : str
        How to compare depth values ("<", "<=", "==", "!=", ">=", ">"). Default
        "<".
    depth_write : bool | None
        Whether the object writes to the depth buffer. With None (default) the
        value depends on the ``alpha_config``.
    alpha_test : float
        The alpha test value for this material. Default 0.0, meaning no alpha
        test is performed.
    alpha_compare : str
        How to compare alpha values ("<", "<=", "==", "!=", ">=", ">"). Default
        "<".
    """

    # Note that in the material classes we define what properties are stored as
    # a uniform, and which are stored on `._store` (and will likely be used as
    # shader templating variables). This can be seen as an abstraction leak, but
    # abstracing it away is inpractical and complex. So we embrace wgpu as the
    # primary rendering backend, and other backends should deal with it.
    # See https://github.com/pygfx/pygfx/issues/272

    uniform_type: ClassVar[dict[str, str]] = dict(
        opacity="f4",
        alpha_test="f4",
        clipping_planes="0*4xf4",  # array<vec4<f32>,3>
    )

    def __init__(
        self,
        *,
        opacity: float = 1,
        clipping_planes: Sequence[ABCDTuple] = (),
        clipping_mode: Literal["ANY", "ALL"] = "ANY",
        alpha_mode: Optional[str] = "auto",
        alpha_config: Optional[dict] = None,
        depth_test: bool = True,
        depth_compare: str = "<",
        depth_write: Literal[None, False, True] = None,
        pick_write: bool = False,
        alpha_test: float = 0.0,
        alpha_compare: str = "<",
        render_queue: Optional[int] = None,
    ):
        super().__init__()

        self._store.uniform_buffer = Buffer(
            array_from_shadertype(self.uniform_type), force_contiguous=True
        )
        self._given_render_queue = 2000  # init to avoid resolving many times
        self._store["use_alpha_test"] = None
        self.opacity = opacity
        self.clipping_planes = clipping_planes
        self.clipping_mode = clipping_mode
        if alpha_config is None:
            self.alpha_mode = alpha_mode or "auto"
        else:
            self.alpha_config = alpha_config
        self.depth_test = depth_test
        self.depth_compare = depth_compare
        self.depth_write = depth_write
        self.pick_write = pick_write
        self.alpha_test = alpha_test
        self.alpha_compare = alpha_compare
        self.render_queue = render_queue  # set last, bc it resolves from other values

    def _set_size_of_uniform_array(self, key: str, new_length: int) -> None:
        """Resize the given array field in the uniform struct if the
        current length does not match the given length. This adjusts
        the uniform type, creates a new buffer from that, and copies
        all data except that of the given field.

        Resetting the uniform buffer will bump the rev and thus trigger
        a pipeline rebuild for all objects that this material is
        attached to.
        """
        current_length = self.uniform_buffer.data[key].shape[0]
        if new_length == current_length:
            return  # early exit

        format = self.uniform_type[key]
        assert "*" in format, f"uniform field {key} '{format}' is not an array"
        n, _, subtype = format.partition("*")
        assert int(n) >= 0

        # Adjust type definition (note that this is originally a class attr)
        self.uniform_type[key] = f"{new_length}*{subtype}"
        # Recreate buffer
        data = self.uniform_buffer.data
        self._store.uniform_buffer = Buffer(
            array_from_shadertype(self.uniform_type), force_contiguous=True
        )
        # Copy data
        for k in data.dtype.names:
            if k != key:
                self.uniform_buffer.data[k] = data[k]

    def _wgpu_get_pick_info(self, pick_value) -> dict:
        """Given a 4 element tuple, sampled from the pick texture,
        return info about what was picked in the object. The first value
        refers to the object id. The use of the remaining values differs
        per material.
        """
        # Note that this is a private friend-method of the renderer.
        return {}

    @property
    def uniform_buffer(self) -> Buffer:
        """The uniform buffer object for this material.

        Properties that are represented in the buffer can be updated cheaply
        (i.e. without requiring shader compilation).
        """
        return self._store.uniform_buffer

    @property
    def opacity(self) -> float:
        """The opacity (a.k.a. alpha value) applied to this material (0..1).

        If the material's color has an alpha smaller than 1, this alpha is multiplied with the opacity.
        """
        return float(self.uniform_buffer.data["opacity"])

    @opacity.setter
    def opacity(self, value: float) -> None:
        value = min(max(float(value), 0), 1)
        self.uniform_buffer.data["opacity"] = value
        self.uniform_buffer.update_full()

    @property
    def clipping_planes(self) -> Sequence[ABCDTuple]:
        """A tuple of planes (abcd tuples) in world space. Points in
        space whose signed distance to the plane is negative are clipped
        (not rendered). Applies to the object to which this material is attached.
        """
        return [
            tuple(float(f) for f in plane.flat)  # type: ignore
            for plane in self.uniform_buffer.data["clipping_planes"]
        ]

    @clipping_planes.setter
    def clipping_planes(self, planes: Sequence[ABCDTuple]):
        if not isinstance(planes, (tuple, list)):
            raise TypeError("Clipping planes must be a list.")
        planes2 = []
        for plane in planes:
            if isinstance(plane, (tuple, list)) and len(plane) == 4:
                planes2.append(plane)
            # maybe someday elif isinstance(plane, linalg.Plane):
            else:
                # Error
                if isinstance(plane, (int, float)) and len(planes) == 4:
                    raise TypeError(
                        "Looks like you passed one clipping plane instead of a list."
                    )
                else:
                    raise TypeError(
                        f"Each clipping plane must be an abcd tuple, not {plane}"
                    )

        # Apply
        # Note that we can create a null-plane using (0, 0, 0, 0)
        self._set_size_of_uniform_array("clipping_planes", len(planes2))
        self._store.clipping_plane_count = len(planes2)
        for i in range(len(planes2)):
            self.uniform_buffer.data["clipping_planes"][i] = planes2[i]
        self.uniform_buffer.update_full()

    @property
    def clipping_plane_count(self) -> int:
        """The number of clipping planes (readonly)."""
        return self._store.clipping_plane_count

    @property
    def clipping_mode(self) -> Literal["ANY", "ALL"]:
        """Set the behavior for multiple clipping planes: "ANY" or "ALL".
        If this is "ANY" (the default) a fragment is discarded
        if it is clipped by any clipping plane. If this is "ALL", a
        fragment is discarded only if it is clipped by *all* of the
        clipping planes.
        """
        return self._clipping_mode

    @clipping_mode.setter
    def clipping_mode(self, value: Literal["ANY", "ALL"]) -> None:
        mode = value.upper()
        if mode in ("ANY", "ALL"):
            self._clipping_mode = mode
        else:
            raise ValueError(f"Unexpected clipping_mode: {value}")

    @property
    def alpha_mode(self) -> AlphaMode:
        """Defines how the the resulting colors are combined with the target color texture.

        The ``material.alpha_mode`` is the recommended way to specify how an
        object's color is combined with the colors of earlier rendered objects.
        It provides preset configurations covering the majority of use-cases.

        The ``alpha_mode`` is a convenient way to set ``alpha_method`` and ``alpha_config`.
        If more control is needed, set ``alpha_config`` directly.

        Modes for method "opaque" (overwrites the value in the output texture):

         * "solid": alpha is ignored.
         * "solid_premul": the alpha is multipled with the color (making it darker).

        Modes for method "blended" (per-fragment blending, a.k.a. compositing):

         * "auto": classic alpha blending, with ``depth_write`` defaulting to True. See note below.
         * "blend": classic alpha blending using the over-operator. ``depth_write`` defaults to False.
         * "add": additive blending that adds the fragment color, multiplied by alpha.
         * "subtract": subtractuve blending that removes the fragment color.
         * "multiply": multiplicative blending that multiplies the fragment color.

        Modes for method "weighted" (order independent blending):

         * "weighted_blend": weighted blended order independent transparency.
         * "weighted_solid": fragments are combined based on alpha, but the final alpha is always 1. Great for e.g. image stitching.

        Modes for method "stochastic" (alpha represents the chance of a fragment being visible):

         * "dither": stochastic transparency with blue noise. This mode handles
           order-independent transparency exceptionally well, but it produces
           results that can look somewhat noisy.
         * "bayer": stochastic transparency with an 8x8 Bayer pattern.

        Note that the special mode "auto" produces reasonable results for common
        use-cases and can handle objects that produce a mix of opaque (alpha=1)
        and transparent (alpha<1) fragments, i.e. it can handle lines, points,
        and text with ``material.aa=True``. Artifacts can occur when objects are rendered out
        of order and/or when objects intersect. A different method such as
        "blend", "dither", or "weighted_blend" is then recommended.

        Note that for methods 'opaque' and 'stochastic', the ``depth_write``
        defaults to True, and for methods 'blended' and 'weighted' the
        ``depth_write`` defaults to False (except when mode is 'auto').

        Note that the value of ``material.alpha_mode`` can be "custom" in case
        ``alpa_config`` is set to a configuration not covered by a preset mode.
        """
        return self._store.alpha_mode

    @alpha_mode.setter
    def alpha_mode(self, alpha_mode: AlphaMode):
        if isinstance(alpha_mode, dict):
            raise TypeError(
                "material.alpha_mode must be a str, not a dict, maybe you meant material.alpha_config?"
            )
        if not isinstance(alpha_mode, str):
            raise TypeError(f"material.alpha_mode must be a str, not {alpha_mode!r}")
        alpha_mode = alpha_mode.lower()
        if alpha_mode in AlphaMethod:
            m = {
                "opaque": "solid",
                "blended": "blend",
                "weighted": "weighted_blend",
                "stochastic": "dither",
            }
            suggestion = m.get(alpha_mode, "solid")
            raise ValueError(
                f"material.alpha_mode is set to {alpha_mode!r} which is an alpha-method. Maybe you meant {suggestion!r}?"
            )
            if alpha_mode not in AlphaMode:
                raise ValueError("Cannot set material.alpha_mode with 'custom'")

        if alpha_mode == "auto":
            d = ALPHA_MODES["blend"].copy()
            d["mode"] = "auto"
        else:
            d = ALPHA_MODES[alpha_mode]
        self.alpha_config = d

    @property
    def alpha_method(self) -> str:
        """The alpha method being used (readonly).

        The alpha method determines the main way how alpha values are used. There are four options:

        * "opaque" means the fragments overwrite value in the color buffer.
        * "blended" means the fragments are blended (composited) with the color buffer.
        * "weighted" means the fragments are blended in a weighted order-independent way.
        * "stochastic" means the fragments are opaque, and alpha represents the chance of a fragment being visible.

        The ``alpha_config`` dictionary specifies extra parameters that affect the behavour of each alpha method.
        The ``alpha_mode`` is a convenient way to set ``alpha_config`` (and ``alpha_method``) using presets

        """
        return self._store.alpha_method

    @property
    def alpha_config(self) -> dict:
        """Dict that defines how the the resulting colors are combined with the target color texture.

        The ``alpha_config`` property is repesented as a dictionary that fully
        describes how the object is combined with other objects rendered at the
        same time. See ``material.alpha_mode`` for convenient presets.

        All possible alpha configurations are grouped in four methods:

        * "opaque": colors simply overwrite the texture, no transparency.
        * "blended": colors are blended with the buffer per-fragment.
        * "weighted": weighted blended order independent transparency, and variants thereof.
        * "stochastic": stochastic transparency, alpha represents the chance of a fragment being visible.

        The ``alpha_config`` dict has at least the following fields:

        * "mode": the corresponding mode or 'custom' (optional when setting).
        * "method": select the alpha-method (mandatory when setting).

        For each method, different options are available, which are represented as
        items in the ``alpha_config`` dict.

        Options for method 'opaque':

        * "premultiply": whether to pre-multiply the color with the alpha values.

        Options for method 'blended':

        * "color_op": the blend operation/equation, any value from ``wgpu.BlendOperation``. Default "add".
        * "color_src": source factor, any value of ``wgpu.BlendFactor``. Mandatory.
        * "color_dst": destination factor, any value of ``wgpu.BlendFactor``. Mandatory.
        * "color_constant": represents the constant value of the constant blend color. Default black.
        * "alpha_op": as ``color_op`` but for alpha.
        * "alpha_src": as ``color_dst`` but for alpha.
        * "alpha_dst": as ``color_src`` but for alpha.
        * "alpha_constant": as ``color_constant`` but for alpha (default 0).

        Options for method 'weighted':

        * "weight"``": the weight factor as wgsl code. Default "alpha", which means use the color's alpha value.
        * "alpha": the used alpha value. Default "alpha", which means use as-is. Can e.g. be set to 1.0
          so that the alpha channel can be used as the weight factor, while the object is otherwise opaque.

        Options for method 'stochastic':

        * "pattern": can be 'blue-noise' for blue noise (default), 'white-noise' for white
          noise, and 'bayer' for a Bayer pattern.
        * "seed": can be 'screen' to have a uniform pattern for the whole screen, 'object' to
          use a per-object seed, and 'element' to have a per-element seed.
          The default is 'element' for  the noise patterns and 'object' for the bayer pattern.


        """
        return self._store.alpha_config.copy()

    @alpha_config.setter
    def alpha_config(self, alpha_config: dict) -> None:
        if not isinstance(alpha_config, dict):
            raise TypeError(
                f"material.alpha_config expects a dict, not {alpha_config!r}"
            )

        # Get the alpha method
        method = alpha_config["method"]
        if method not in AlphaMethod:
            raise ValueError(
                f"Invalid 'method' in material.alpha_config: {method!r}, expected one of {AlphaMethod}"
            )

        # Init with a preset?
        mode = alpha_config.get("mode", None)
        if mode and mode not in ("auto", "custom"):
            if mode not in ALPHA_MODES:
                raise ValueError(f"Invalid mode in material.alpha_config: {mode!r}")
            d = ALPHA_MODES[mode]
            if d["method"] != method:
                raise ValueError(f"Invalid mode {mode!r} for method {method!r}.")
            alpha_config = {**d, **alpha_config}

        if method == "opaque":
            keys = ["premultiply_alpha"]
            defaults = {}
        elif method == "stochastic":
            keys = ["pattern", "seed"]
            defaults = {"pattern": "blue_noise", "seed": "screen"}
        elif method == "blended":
            keys = [
                "color_op",
                "color_src",
                "color_dst",
                "color_constant",
                "alpha_op",
                "alpha_src",
                "alpha_dst",
                "alpha_constant",
            ]
            defaults = {
                "color_op": "add",
                "alpha_op": "add",
                "color_constant": (0, 0, 0),
                "alpha_constant": 0,
            }
        elif method == "weighted":
            keys = ["weight", "alpha"]
            defaults = {"weight": "alpha", "alpha": "alpha"}
        else:
            assert False, "if-tree is missing a case"  # noqa: B011

        # Get options. This has the same form as the values in ALPHA_MODES, so they can be compared
        options = self._get_alpha_config_options(method, keys, defaults, alpha_config)

        # Derive mode
        the_mode = "custom"
        for mode_name, mode_dict in ALPHA_MODES.items():
            if mode_dict["method"] == method:
                if mode_dict == options:
                    the_mode = mode_name
                    break
        if mode == "auto" and the_mode == "blend":
            the_mode = "auto"

        # Save
        self._store.alpha_mode = the_mode
        self._store.alpha_method = method
        self._store.alpha_config = ReadOnlyDict(
            {
                "mode": the_mode,
                "method": method,
                **options,
            }
        )
        self._derive_render_queue()

    @property
    def render_queue(self) -> int:
        """An integer that represents the group that the renderer uses to sort objects.

        The property is intended for advanced usage; by default it is determined automatically
        based on ``alpha_mode``, ``alpha_method`` and ``alpha_test``.

        The ``render_queue`` is a number between 1 and 4999. The builtin values are:

        * 1000: background.
        * 2000: opaque non-blending objects.
        * 2400: opaque objects with a discard based on alpha (i.e. using ``alpha_test`` or "stochastic" alpha-method).
        * 2600: objects with alpha-mode 'auto'.
        * 3000: transparent objects.
        * 4000: overlay.

        Objects with ``render_queue`` between 1501 and 2500 are sorted front-to-back. Otherwise
        objects are sorted back-to-front.

        It can be tempting to use ``material.render_queue`` to control the render order of individual objects,
        but for that purpose ``ob.render_order`` is more appropriate.
        """
        return self._render_queue

    @render_queue.setter
    def render_queue(self, render_queue: int | None):
        render_queue = int(render_queue or 0)
        if render_queue < 0 or render_queue > 4999:
            if render_queue == -1:
                render_queue = 0
            else:
                raise ValueError(
                    "Material.render_queue must be an int between 1 and 4999, or 0 for 'default'."
                )
        self._given_render_queue = render_queue
        self._derive_render_queue()

    @property
    def render_queue_is_set(self):
        """Whether the ``render_queue`` property is set. Otherwise it's auto-determined."""
        return bool(self._given_render_queue)

    def _derive_render_queue(self):
        if self._given_render_queue:
            render_queue = self._given_render_queue
        elif self._store.alpha_mode == "auto":
            # An "opaque" auto-object, but it may have semi-transparent fragments,
            # because of aa, or maps with transparent regions. So back-to-front.
            render_queue = 2600
        else:
            alpha_method = self._store.alpha_method
            if alpha_method == "opaque":
                if not self._store.use_alpha_test:
                    render_queue = 2000
                else:
                    render_queue = 2400
            elif alpha_method == "stochastic":
                render_queue = 2400
            else:  # alpha_method in ["blended", "weighted"]
                render_queue = 3000
        self._render_queue = render_queue

    def _get_alpha_config_options(
        self, method: str, keys: list, default_dict: dict, given_dict: dict
    ):
        err_preamble = f"material.alpha_config for method {method!r}"
        assert isinstance(given_dict, dict)
        source_dict = given_dict.copy()
        result_dict = {"method": method, **default_dict}
        for key in ["method", "mode"]:
            source_dict.pop(key, None)
        try:
            for key in keys:
                if key in source_dict:
                    result_dict[key] = source_dict.pop(key)
        except KeyError as err:
            raise KeyError(f"{err_preamble} is missing field {err.args[0]!r}") from None
        if source_dict:
            raise ValueError(f"{err_preamble} contains invalid fields: {source_dict!r}")
        return result_dict

    @property
    def depth_test(self) -> bool:
        """Whether the object takes the depth buffer into account.

        When set to True, the fragment's depth is tested using ``depth_compare`` against the depth buffer.
        """
        return self._store.depth_test

    @depth_test.setter
    def depth_test(self, value: bool) -> None:
        self._store.depth_test = bool(value)

    @property
    def depth_compare(self):
        """The way to compare the depth with the value in the buffer.

        Possible values are "<", "<=", "==", "!=", ">=", ">". Default "<".
        Note that this only applies if ``depth_test`` is set to True.
        """
        return self._store.depth_compare

    @depth_compare.setter
    def depth_compare(self, value) -> None:
        if not (isinstance(value, str) and value in TEST_COMPARE_VALUES):
            raise TypeError(
                "Material.depth_compare must be a str in {TEST_COMPARE_VALUES!r}, not {value!r}"
            )
        self._store.depth_compare = value

    @property
    def depth_write(self) -> bool:
        """Whether this material writes to the depth buffer, preventing other objects being drawn behind it.

        Can be set to:

        * True: yes, write depth.
        * False: no, don't write depth.
        * None: auto-determine based on ``alpha_coalpha_mode`` and ``alpha_method`` (default).

        The auto-option provides good default behaviour for common use-case, but
        if you know what you're doing you should probably just set this value to
        True or False.
        """
        depth_write = self._store.depth_write
        if depth_write is None:
            depth_write = (
                self._store.alpha_mode == "auto"
                or self._store.alpha_method in ("opaque", "stochastic")
            )
        return depth_write

    @depth_write.setter
    def depth_write(self, value: Literal[None, False, True]) -> None:
        if value is None:
            self._store.depth_write = None
        elif isinstance(value, bool):
            self._store.depth_write = bool(value)
        else:
            raise TypeError("material.depth_write must be bool or None.")

    @property
    def depth_write_is_set(self):
        """Whether the ``depth_write`` property is set. Otherwise it's auto-determined."""
        return self._store.depth_write is not None

    @property
    def alpha_test(self) -> float:
        """The alpha test value for this material.

        When ``alpha_test`` is set to a value > 0, the fragment is discarded if ``alpha < alpha_test``.
        This is useful for e.g. grass or foliage textures, where the texture has a lot of transparent
        areas. Also see ``alpha_compare``.
        """
        return self.uniform_buffer.data["alpha_test"]

    @alpha_test.setter
    def alpha_test(self, value: float) -> None:
        value = min(max(float(value), 0), 1)
        self.uniform_buffer.data["alpha_test"] = value
        self.uniform_buffer.update_full()
        # Store whether the alpha test is active, so we can invalidate shaders
        use_alpha_test = value != 0
        if use_alpha_test != self._store.use_alpha_test:
            self._store.use_alpha_test = use_alpha_test
            self._derive_render_queue()

    @property
    def alpha_compare(self) -> str:
        """The way to compare the alpha value.

        Possible values are "<", "<=", "==", "!=", ">=", ">". Default "<".
        Note that this only applies if the alpha test is performed (i.e. ``alpha_test`` is nonzero).
        """
        return self._store.alpha_compare

    @alpha_compare.setter
    def alpha_compare(self, value: str) -> None:
        if not (isinstance(value, str) and value in TEST_COMPARE_VALUES):
            raise TypeError(
                "Material.alpha_compare must be a str in {TEST_COMPARE_VALUES!r}, not {value!r}"
            )
        self._store.alpha_compare = value

    @property
    def _gfx_use_alpha_test(self) -> bool:
        """For internal use; if the alpha test is used, the shader must discard, which we don't want to do unless needed."""
        return self._store.use_alpha_test

    @property
    def pick_write(self) -> bool:
        """Whether this material is picked by the pointer."""
        return self._store.pick_write

    @pick_write.setter
    def pick_write(self, value: int) -> None:
        if not isinstance(value, (bool, int)):
            raise TypeError("Material.pick_write must be bool.")
        self._store.pick_write = bool(value)



import weakref

from ..utils.trackable import Trackable


class Resource(Trackable):
    """Base class for :class:`~pygfx.resources.Buffer` and :class:`~pygfx.resources.Texture`."""

    _resource_counts = {}  # Just to track the number of buffers and textures alive
    _rev = 0  # integer hash

    def __init__(self):
        super().__init__()
        cname = self.__class__.__name__
        Resource._resource_counts[cname] = Resource._resource_counts.get(cname, 0) + 1

    def __del__(self):
        cname = self.__class__.__name__
        Resource._resource_counts[cname] -= 1

    def _gfx_mark_for_sync(self):
        resource_update_registry._gfx_mark_for_sync(self)

    @property
    def rev(self) -> int:
        """The revision number (integer).

        The number changes when ``update_range()`` is called. The number is
        monotonically increasing and globally unique (no two buffers/textures
        have the same rev). This makes that it can be used as hash for the data
        content.
        """
        return self._rev


class ResourceUpdateRegistry:
    """Singleton registry to keep track of resources that need to be updated for the wgpu-backend."""

    def __init__(self):
        self._syncable = weakref.WeakSet()

    def _gfx_mark_for_sync(self, resource):
        """Register the given resource for synchronization. Only adds the resource
        if it's wgpu-counterpart already exists, and when it has pending uploads.
        """
        if not isinstance(resource, Resource):
            raise TypeError("Given object is not a Resource")
        if resource._wgpu_object is not None:
            self._syncable.add(resource)

    def get_syncable_resources(self, *, flush=False):
        """Get the set of resources that need syncing. If setting flush
        to True, the caller is responsible for syncing the resources.
        """
        syncable = set(self._syncable)
        if flush:
            self._syncable.clear()
        return syncable


resource_update_registry = ResourceUpdateRegistry()



import wgpu  # only for flags/enums

from ....objects import Text
from ....materials import TextMaterial
from ....utils.text._shaper import REF_GLYPH_SIZE

from .. import (
    register_wgpu_render_function,
    load_wgsl,
    BaseShader,
    Binding,
    GfxSampler,
    GfxTextureView,
)


@register_wgpu_render_function(Text, TextMaterial)
class TextShader(BaseShader):
    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)
        material = wobject.material
        self["is_screen_space"] = wobject.screen_space
        self["is_multi_text"] = wobject.is_multi
        self["aa"] = material._gfx_effective_aa
        self["REF_GLYPH_SIZE"] = REF_GLYPH_SIZE

    def get_bindings(self, wobject, shared):
        geometry = wobject.geometry
        material = wobject.material

        sbuffer = "buffer/read_only_storage"
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
            Binding("s_positions", sbuffer, geometry.positions, "VERTEX"),
        ]

        tex = shared.glyph_atlas_texture
        sampler = GfxSampler("linear", "clamp")
        tex_view = GfxTextureView(tex)
        bindings.append(Binding("s_atlas", "sampler/filtering", sampler, "FRAGMENT"))
        bindings.append(Binding("t_atlas", "texture/auto", tex_view, "FRAGMENT"))

        # Let the shader generate code for our bindings
        bindings = {i: b for i, b in enumerate(bindings)}
        self.define_bindings(0, bindings)

        bindings1 = {}
        bindings1[0] = Binding(
            "s_glyph_info", sbuffer, shared.glyph_atlas_info_buffer, "VERTEX"
        )
        bindings1[1] = Binding("s_glyph_data", sbuffer, geometry.glyph_data, "VERTEX")

        return {
            0: bindings,
            1: bindings1,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_list,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        n = wobject.geometry.glyph_data.draw_range[1] * 6
        return {
            "indices": (n, 1),
        }

    def get_code(self):
        return load_wgsl("text.wgsl")



from math import floor, ceil

import numpy as np

from ._base import Resource
from ._utils import (
    get_element_format_from_numpy_array,
    calculate_texture_chunk_size,
    get_merged_blocks_from_mask_3d,
    check_data_is_clean_for_performance,
    is_little_endian,
    make_little_endian,
    logger,
)


class Texture(Resource):
    """The Texture represents structured 1D, 2D or 3D data on the GPU.

    A texture can be used to represent e.g. image data or colormaps. They can
    also serve as a render target (for the renderer). Supports texture stacks,
    cube textures, and mipmapping.

    Parameters
    ----------
    data : array | None
        The initial data of the texture. It must support the buffer-protocol,
        (e.g. a bytes or numpy array). If None, ``size`` and ``format`` must be
        provided. The data will be accessible at ``texture.data``, no copies are
        made. The dtype must be compatible with wgpu texture formats.
    dim : int
        The dimensionality of the array (1, 2 or 3).
    size : tuple | None
        The extent ``(width, height, depth)`` of the array. If None, it is
        derived from `dim` and the shape of the data. The texture can also
        represent a stack of images by setting ``dim=2`` and ``depth > 1``, or a
        cube image by setting ``dim=2`` and ``depth==6``.
    format : None | str | ElementFormat | wgpu.TextureFormat
        A format string describing the pixel/voxel format. This can follow
        pygfx' ``ElementFormat`` e.g. "1xf4" for intensity, "3xu1" for rgb, etc.
        Can also be wgpu's ``TextureFormat``. Optional: if None, it is
        automatically determined from the data.
    colorspace : str
        If this data is used as color, it is interpreted to be in this
        colorspace. Can be "srgb", "tex-srgb", "physical", "yuv420p", or "yuv444p". Default "srgb".
    colorrange : str
        For YUV textures, this is either "limited", or "full". For the limited range,
        the luma plane is limited between 16-235, and the chroma planes (U and V) are
        limited to 16-240. While it may seem suboptimal, many videos are stored in
        the limited colorrange.
    generate_mipmaps : bool
        If True, automatically generates mipmaps when transferring data to the
        GPU. Default False.
    chunk_size : None | tuple | int
        The chunk size to use for uploading data to the GPU, expressed in elements (not bytes).
        When None (default) an optimal chunk size is determined automatically.
        A 3-tuple can be given to provide a size for each dimension, or an integer
        to apply for all dimensions.
    force_contiguous : bool
        When set to true, the texture goes into a stricter mode, forcing set data
        to be c_contiguous. This ensures optimal upload performance for cases when
        the data changes often.
    usage : int | wgpu.TextureUsage
        The wgpu ``usage`` flag for this texture. Optional: typically pygfx can
        derive how the texture is used and apply the appropriate flag. In cases
        where it doesn't this param provides an override. This is a bitmask flag
        (values are OR'd).

    Performance tips:

    * If the given data is not c_contiguous, extra memory-copies may be needed
      at upload time, which reduces performance when the data is changed often.
    * RGB textures do not exist in wgpu, but are emulated with an RGBA texture.
      This may introduce extra memory copies, which reduces performance when
      data is changed often.
    * Setting ``force_contiguous`` ensures that the set data is contiguous and
      not RGB, it is recommended to use this when the texture data is dynamic.
    """

    def __init__(
        self,
        data=None,
        *,
        dim,
        size=None,
        format=None,
        colorspace="srgb",
        colorrange="limited",
        generate_mipmaps=False,
        chunk_size=None,
        force_contiguous=False,
        usage=0,
    ):
        super().__init__()
        Resource._rev += 1
        self._rev = Resource._rev

        # Attributes for internal use, updated by other parts of pygfx.
        self._wgpu_object = None
        self._wgpu_usage = int(usage)
        self._wgpu_mip_level_count = 1
        self._wgpu_emulate_rgb = False

        # Init
        self._data = None
        self._view = None
        self._force_contiguous = bool(force_contiguous)
        assert dim in (1, 2, 3)
        self._store.dim = int(dim)
        self._colorspace = (colorspace or "srgb").lower()
        assert self._colorspace in (
            "srgb",
            "tex-srgb",
            "physical",
            "yuv420p",
            "yuv444p",
        )
        self._colorrange = (colorrange or "limited").lower()
        assert self._colorrange in ("limited", "full")
        self._generate_mipmaps = bool(generate_mipmaps)

        # Normalize size
        if size is None:
            pass
        elif len(size) == 2:
            size = int(size[0]), int(size[1]), 1
        elif len(size) == 3:
            size = int(size[0]), int(size[1]), int(size[2])
        else:
            raise ValueError("Texture size must be a 2-tuple or 3-tuple or None.")

        # Process data
        if data is not None:
            # Store data and view, and do some basic checks.
            # The view is a numpy array, but we go via memoryview to ensure data follows the buffer protocol.
            self._data = data
            self._view = view = np.asarray(memoryview(data))
            if self._force_contiguous:
                check_data_is_clean_for_performance("texture", view)
            the_nbytes = view.nbytes
            # Establish texture size
            if size is not None:
                the_size = size
            else:
                the_size = size_from_array(view, dim)
            reshape_array(view, the_size)
            # Establish format
            element_format = get_element_format_from_numpy_array(view)
            if element_format is None:
                raise ValueError(
                    f"Unsupported dtype/format for texture data: {view.dtype}"
                )
            nchannels = int(np.prod(view.shape[3:], initial=1))
            if not (1 <= nchannels <= 4):
                raise ValueError(
                    f"Expected 1-4 texture color channels, got {nchannels}."
                )
            if nchannels == 3:
                self._wgpu_emulate_rgb = True
                if self._force_contiguous:
                    raise ValueError(
                        "When force_contiguous is set, the texture data cannot be rgb, because it requires a padding operation on upload."
                    )
            detected_format = (f"{nchannels}x" + element_format).lstrip("1x")
        elif size is not None and format is not None:
            the_size = size
            the_nbytes = 0
            detected_format = None
        else:
            raise ValueError(
                "Texture must be instantiated with either data or size and format."
            )

        # Check size
        if not all(the_size[i] > 0 for i in range(3)):
            raise ValueError("Texture size cannot be zero.")

        # Store derived props
        self._store.nbytes = the_nbytes
        self._store.size = the_size
        if format is not None:
            self._store.format = str(format)
        elif detected_format:
            self._store.format = detected_format
        else:
            self._store.format = None

        # Get optimal chunk size
        if the_nbytes == 0:  # data is None or empty
            chunk_size = (0, 0, 0)
        elif chunk_size is None:
            chunk_size = calculate_texture_chunk_size(
                the_size,
                bytes_per_element=the_nbytes // int(np.prod(the_size)),
                byte_align=16,
                target_chunk_count=32,
            )
        else:
            if isinstance(chunk_size, int):
                chunk_size = chunk_size, chunk_size, chunk_size
            elif not len(chunk_size) == 3:
                raise ValueError("Chunk size must be int or tuple of 3 ints.")
            chunk_size = tuple(
                min(max(int(chunk_size[i]), 1), the_size[i]) for i in range(3)
            )

        # Init chunks map
        if data is None:
            self._chunks_dirt_flag = 0
            self._chunk_size = (0, 0, 0)
            self._chunk_mask = None
            self._chunk_list = []
        elif the_nbytes == 0:
            self._chunks_dirt_flag = 0
            self._chunk_size = (0, 0, 0)
            self._chunk_mask = np.ones((0, 0, 0), bool)
            self._chunk_list = None
        else:
            self._chunks_dirt_flag = 2
            self._chunk_size = chunk_size
            shape = tuple(ceil(the_size[i] / self._chunk_size[i]) for i in (2, 1, 0))
            self._chunk_mask = np.ones(shape, bool)
            self._chunk_list = None

    @property
    def dim(self):
        """The dimensionality of the texture (1, 2, or 3)."""
        return self._store.dim

    @property
    def data(self):
        """The data for this texture.

        Can be None if the data only exists on the GPU. This object is the same
        that was given to instantiate this object or with ``set_data()``.
        """
        return self._data

    @property
    def view(self):
        """A numpy array view on the data of this texture.

        Can be None if the data only exists on the GPU. This is a view on the
        same memory as ``.data``. It's ``.shape[:3]`` matches ``reversed(size)``.
        """
        return self._view

    @property
    def nbytes(self):
        """Get the number of bytes in the texture."""
        return self._store.nbytes

    @property
    def size(self):
        """The size of the texture as (width, height, depth).
        (always a 3-tuple, regardless of the dimension).
        """
        return self._store.size

    @property
    def format(self):
        """The texture format.

        Usually a pygfx format specifier (e.g. 'u2' for scalar uint16, or '3xf4'
        for RGB float32), but can also be a value from ``wgpu.TextureFormat``.
        """
        return self._store.format

    @property
    def usage(self):
        """Bitmask indicating how the texture can be used in a wgpu pipeline."""
        return self._wgpu_usage

    @property
    def colorspace(self):
        """If this data is used as color, it is interpreted to be in this colorspace.
        Can be "srgb", "tex-srgb", "physical", "yuv420p", or "yuv444p". Default "srgb".

        * "srgb": the data represents intensity, rgb, or rgba pixels in the sRGB space.
          sRGB is a standard color space designed for consistent representation of colors
          across devices like monitors. Most images store colors in this space.
          The shader convers sRGB colors to physical in the shader before doing color computations.
        * "tex-srgb": the underlying texture will be of an sRGB format. This means the data
          is automatically converted to sRGB when it is sampled. This results in better glTF
          compliance (because interpolation in the sampling happens in linear space).
          Note that sampling *always* results in the sRGB values, also when not interpreted as color.
          Only supported for rgb and rgba data.
        * "physical": the colors are (already) in the physical / linear space, where lighting
          calculations can be applied. Shader code that interprets the data as color will use it as-is.
        * "yuv420p": A common video format. The data is represented as 3 planes (y, u, and v).
          The y represents intensity, and is at full resolution. The u and v planes are a
          quarter of the size. The planes must be stored in two layers of the texture,
          with the u and v plane next to each-other in top half the second layer.
        * "yuv444p": A lesser common video format. The data is represented as 3 planes
          (y, u, and v) similar to yuv420p however the u and v planes are stored
          at full resolution.
        """
        return self._colorspace

    @property
    def colorrange(self):
        """For YUV textures, this is either "limited", or "full".

        * "limited": The luma plane (Y) is limited to the range of 16-235 for 8 bits.
                     The chroma planes (U and V) are limited to the range of 16-240 for 8 bits
        * "full": The luma plane and chroma plane use the full range of the storage format.

        See the following links from the FFMPEG documentation for more details:
        https://trac.ffmpeg.org/wiki/colorspace
        https://ffmpeg.org/doxygen/7.0/pixfmt_8h_source.html#l00609
        """
        return self._colorrange

    @property
    def generate_mipmaps(self):
        """Whether to automatically generate mipmaps when uploading to the GPU."""
        return self._generate_mipmaps

    def send_data(self, offset, data):
        """Send a chunk of data to the GPU.

        This provides a way to upload data to textures that don't have local
        data (i.e. ``texture.data is None``). It is intended for use-cases where
        data-copies must be avoid for performance. Can only be used when the
        texture has no local data, and requires ``usage=wgpu.TextureUsage.COPY_DST``.

        The offset must be a (width, height, depth) tuple.
        Note that in contrast to the ``update_x`` methods, multiple calls are
        not combined; each call to ``send_data()`` results in one upload
        operation.

        Example:

        .. code-block:: py

            tex = gfx.Texture(
                size=(64, 64, 1),
                dim=2,
                format=wgpu.TextureFormat.rgba8unorm,
                usage=wgpu.TextureUsage.COPY_DST,
                force_contiguous=True,
            )
        """
        # Check data
        if self._data is not None:
            raise RuntimeError(
                "Can only use texture.send_data() if the texture has no local data."
            )
        nchannels = nchannels_from_format(self.format)
        if nchannels > 1:
            if data.shape[-1] != nchannels:
                raise ValueError(
                    f"Expected {nchannels} channels with format '{self.format}', but shape[-1] is {data.shape[-1]}."
                )
        # Check offset
        if not isinstance(offset, (list, tuple)) and len(offset) == 3:
            raise ValueError("Offset must be a tuple of 3 ints")
        offset = tuple(int(i) for i in offset)
        if any(b < 0 for b in offset):
            raise ValueError("offset must not be negative")
        # Get data size
        shape = list(data.shape)
        need_shape_length = 3 if nchannels == 1 else 4
        while len(shape) < need_shape_length:
            shape.insert(0, 1)
        size = tuple(reversed(shape[:3]))
        # Check whether it fits
        if any(
            (o1 + s1) > s2 for o1, s1, s2 in zip(offset, size, self.size, strict=True)
        ):
            raise ValueError("The data with this offset does not fit.")
        # Create chunk
        data = np.asarray(data).view()
        data.shape = shape
        if not data.flags.c_contiguous:
            if self._force_contiguous:
                raise ValueError(
                    "When force_contiguous is set, data passed to send_data() must be contiguous."
                )
            data = np.ascontiguousarray(data)
        self._chunk_list.append((offset, size, data))
        # Request sync
        Resource._rev += 1
        self._rev = Resource._rev
        self._gfx_mark_for_sync()

    def set_data(self, data):
        """Reset the data to a new array.

        This avoids a data-copy compared to doing ``texture.data[:] = new_data``.
        The new data must fit the teture's size and format.
        """
        # Get view
        view = np.asarray(memoryview(data))
        # Do couple of checks
        if self._force_contiguous:
            check_data_is_clean_for_performance("texture", view)
        if view.nbytes != self._view.nbytes:
            raise ValueError("texture.set_data() nbytes does not match.")
        if view.dtype != self._view.dtype:
            raise ValueError("texture.set_data() format does not match.")
        # Make sure the shape is ok.
        reshape_array(view, self.size)
        # Ok
        self._data = data
        self._view = view
        self.update_full()

    def update_full(self):
        """Mark the whole data for upload."""
        if self._chunks_dirt_flag == 2:
            return
        # self._chunk_mask.fill(True)  # implicit
        self._chunks_dirt_flag = 2
        Resource._rev += 1
        self._rev = Resource._rev
        self._gfx_mark_for_sync()

    def update_indices(self, indices_x, indices_y, indices_z):
        """Mark specific indices for upload.

        The given arrays represent the indices for x, y, and z, respectively.
        So they must be equal in length, similar to what ``np.where()`` returns.
        They can also be None, to indicate the full range, like ":" does with slicing.
        """
        if self._chunks_dirt_flag == 2:
            return
        full_slice = slice(None, None, None)
        div = self._chunk_size
        indices_x = (
            full_slice if indices_x is None else (np.asarray(indices_x) // div[0])
        )
        indices_y = (
            full_slice if indices_y is None else (np.asarray(indices_y) // div[1])
        )
        indices_z = (
            full_slice if indices_z is None else (np.asarray(indices_z) // div[2])
        )
        self._chunk_mask[indices_z, indices_y, indices_x] = True
        self._chunks_dirt_flag = 1
        Resource._rev += 1
        self._rev = Resource._rev
        self._gfx_mark_for_sync()

    def update_range(self, offset, size):
        """Mark a certain range of the data for upload to the GPU.
        The offset and (sub) size should be (width, height, depth)
        tuples. Numpy users beware that an arrays shape is (height, width)!
        """
        if self._chunks_dirt_flag == 2:
            return
        full_size = self.size
        # Check input
        assert isinstance(offset, tuple) and len(offset) == 3
        assert isinstance(size, tuple) and len(size) == 3
        if any(s == 0 for s in size):
            return
        elif any(s < 0 for s in size):
            raise ValueError("Update size must not be negative")
        elif any(b < 0 for b in offset):
            raise ValueError("Update offset must not be negative")
        # offset + size being larger than the full size is tolaterated.
        # Get indices
        div = self._chunk_size
        index_a = tuple(floor(offset[i] / div[i]) for i in range(3))
        index_b = tuple(
            ceil(min(full_size[i], offset[i] + size[i]) / div[i]) for i in range(3)
        )
        # Update map
        self._chunk_mask[
            index_a[2] : index_b[2], index_a[1] : index_b[1], index_a[0] : index_b[0]
        ] = True
        self._chunks_dirt_flag = 1
        Resource._rev += 1
        self._rev = Resource._rev
        self._gfx_mark_for_sync()

    def _gfx_get_chunk_descriptions(self):
        """Get a list of (offset, size) tuples, that can be
        used in _gfx_get_chunk_data(). This method also clears
        the chunk dirty statuses.
        """

        # In no-local-data mode, we (only) have a chunk list
        if self._chunk_list:
            chunks = self._chunk_list
            self._chunk_list = []
            return chunks

        if not self._chunks_dirt_flag:
            return []
        elif self._chunks_dirt_flag == 2:
            chunk_descriptions = [((0, 0, 0), self.size)]
        elif np.all(self._chunk_mask):
            chunk_descriptions = [((0, 0, 0), self.size)]
        else:
            # Get merged chunk blocks, using a smart algorithm.
            chunk_blocks = get_merged_blocks_from_mask_3d(self._chunk_mask)

            # Turn into proper descriptions, with chunk indices/counts scaled with the chunk size.
            chunk_descriptions = []
            chunk_size = self._chunk_size
            full_size = self._store["size"]
            for block in chunk_blocks:
                offset = (
                    block.x * chunk_size[0],
                    block.y * chunk_size[1],
                    block.z * chunk_size[2],
                )
                size = (
                    min(block.nx * chunk_size[0], full_size[0] - offset[0]),
                    min(block.ny * chunk_size[1], full_size[1] - offset[1]),
                    min(block.nz * chunk_size[2], full_size[2] - offset[2]),
                )
                chunk_descriptions.append((offset, size))

        # Reset
        self._chunks_dirt_flag = 0
        self._chunk_mask.fill(False)

        return chunk_descriptions

    def _gfx_get_chunk_data(self, offset, size, pad_value=0.0):
        """Return subdata as a contiguous array."""
        full_size = self._store["size"]
        emulate_rgb = self._wgpu_emulate_rgb

        # Get chunk
        if offset == (0, 0, 0) and size == full_size and not emulate_rgb:
            chunk = self._view
        else:
            # Calculate slice
            slice_per_dim = []
            for d in reversed(range(3)):
                slice_per_dim.append(slice(offset[d], offset[d] + size[d]))
            slice_per_dim = tuple(slice_per_dim)
            # Slice, pad, make contiguous. Note that the chance of the chunks not being contiguous is pretty big.
            chunk = self._view[slice_per_dim]
            if emulate_rgb:
                if self._force_contiguous:
                    logger.warning(
                        "force_contiguous was set, but still need pixel padding"
                    )
                # Copy into array with one extra channel
                nchannels = chunk.shape[3]  # usually 3 for rgb, padded to make rgba
                padded_shape = (*chunk.shape[:3], nchannels + 1)
                padded_chunk = np.full(padded_shape, pad_value, dtype=chunk.dtype)
                padded_chunk[:, :, :, :nchannels] = chunk
                chunk = padded_chunk

        # Normalize chunk
        if not is_little_endian(chunk):
            chunk = make_little_endian(chunk)
            if self._force_contiguous:
                logger.warning(
                    "force_contiguous was set, but chunk data is still big endian"
                )
        elif not chunk.flags.c_contiguous:
            # Easily happens when slicing in anything but the last dimension
            chunk = np.ascontiguousarray(chunk)

        return chunk


def nchannels_from_format(format):
    if len(format) == 2 and format[0] in "usf":
        return 1  # e.g. u1, s2, f4
    elif format[0].isnumeric() and format[1] == "x":
        return int(format.split("x")[0])
    elif format[0] == "r":
        if format[1] != "g":
            return 1
        elif format[2] != "b":
            return 2
        elif format[3] != "a":
            return 3
        else:
            return 4
    elif format.startswith("bgra"):
        return 4
    else:
        # Other prefixes: astc, bc, eac, etc2, depth, stencil
        raise ValueError(f"Cannot derive nchannels from format '{format}' yet.")


def size_from_array(data, dim):
    # Check if shape matches dimension
    shape = data.shape

    if len(shape) not in (dim, dim + 1):
        raise ValueError(
            f"Can't map shape {shape} on {dim}D tex. Maybe also specify size?"
        )
    # Determine size based on dim and shape
    if dim == 1:
        return shape[0], 1, 1
    elif dim == 2:
        return shape[1], shape[0], 1
    else:  # dim == 3:
        return shape[2], shape[1], shape[0]


def reshape_array(view, size):
    """Reshape array so it's shape[0:3] match size."""
    expected_shape = tuple(reversed(size))
    if expected_shape != view.shape[:3]:
        # This can fail if the data is not contiguous and strides don't work out.
        view.shape = (*expected_shape, -1)



from collections import defaultdict
from enum import Enum
from time import perf_counter
from typing import Union, Optional
from weakref import ref

from rendercanvas.base import log_exception


CLICK_DEBOUNCE = 0.5  # in seconds


class EventType(str, Enum):
    # Keyboard
    KEY_DOWN = "key_down"
    KEY_UP = "key_up"
    # Pointer
    POINTER_DOWN = "pointer_down"
    POINTER_MOVE = "pointer_move"
    POINTER_UP = "pointer_up"
    POINTER_ENTER = "pointer_enter"
    POINTER_LEAVE = "pointer_leave"
    CLICK = "click"
    DOUBLE_CLICK = "double_click"
    # Wheel
    WHEEL = "wheel"
    # Window
    CLOSE = "close"
    RESIZE = "resize"


class Event:
    """Event base class.

    If a target is set, an event can bubble up through a hierarchy
    of targets, connected through a ``parent`` property.
    To prevent an event from bubbling up, use ``stop_propagation``.

    It is also possible to cancel events, which will stop any further
    handling of the event (also by the same target).

    Parameters
    ----------
    type : Union[str, EventType]
        The name of the event.
    bubbles : bool
        If True, the event bubbles up through the scene tree.
    target : EventTarget
        The object onto which the event was dispatched.
    root : RootEventHandler
        A reference to the root event handler.
    time_stamp : float
        The time at which the event was created (in seconds). Might not be an actual
        time stamp so please only use this for relative time measurements.
    cancelled : bool
        A boolean value indicating whether the event is cancelled.
    event_type : str
        Unused.

    """

    def __init__(
        self,
        type: Union[str, EventType],
        *,
        bubbles=True,
        target: "EventTarget" = None,
        root: "RootEventHandler" = None,
        time_stamp: Optional[float] = None,
        cancelled: bool = False,
        # Swallow event_type to ease conversion from wgpu events to Event objects
        event_type: Optional[str] = None,
    ):
        self._type = type
        # Using perf_counter_ns instead of perf_counter
        # should give us a bit more accuracy
        self._time_stamp = time_stamp or perf_counter()
        self._bubbles = bubbles
        self._target = target
        self._current_target = target
        self._cancelled = cancelled
        self._root = root

    @property
    def type(self) -> str:
        """A string representing the name of the event."""
        return self._type

    @property
    def root(self) -> "RootEventHandler":
        """A reference to the root event handler."""
        return self._root

    @property
    def time_stamp(self) -> float:
        """The time at which the event was created (in seconds). Might not be
        an actual time stamp so please only use this for relative time measurements."""
        return self._time_stamp

    @property
    def bubbles(self) -> bool:
        """A boolean value indicating whether or not the event bubbles up through
        the scene tree."""
        return self._bubbles

    @property
    def target(self) -> "EventTarget":
        """The object onto which the event was dispatched."""
        return self._target

    @property
    def current_target(self) -> "EventTarget":
        """The object that is currently handling the event. During event bubbling
        this property will be updated whenever an event bubbles up in the hierarchy."""
        return self._current_target

    @property
    def cancelled(self) -> bool:
        """A boolean value indicating whether the event is cancelled."""
        return self._cancelled

    def stop_propagation(self):
        """Stops propagation of events further along in the scene tree."""
        self._bubbles = False

    def cancel(self):
        """Cancels the event and stops propagation."""
        self._cancelled = True
        self.stop_propagation()

    def _retarget(self, target):
        self._target = target
        self._update_current_target(target)

    def _update_current_target(self, target):
        self._current_target = target


class KeyboardEvent(Event):
    """Keyboard button press.

    Parameters
    ----------
    args : Any
        Positional arguments are forwarded to the :class:`base class
        <pygfx.objects.Event>`.
    key : str
        The key that was pressed.
    modifiers : tuple
        The modifiers that were pressed while the key was pressed.
    kwargs : Any
        Additional keyword arguments are forward to the :class:`base class
        <pygfx.objects.Event>`.

    """

    def __init__(self, *args, key, modifiers=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.key = key
        self.modifiers = modifiers or ()


class PointerEvent(Event):
    """Mouse/Touch Event.

    Parameters
    ----------
    args : Any
        Positional arguments are forwarded to the :class:`base class
        <pygfx.objects.Event>`.
    x : int
        The x position of the cursor or touch in screen space (px).
    y : int
        The y position of the cursor or touch in screen space (px).
    button : int
        The integer value of the button being pushed.
    buttons : tuple
        The list of string name of the buttons being pushed.
    modifiers : tuple
        The modifiers that were pressed while the key was pressed.
    ntouches : int
        The total number of synchronous touches.
    touches : list
        A list of all currently occurring touches.
    pick_info : dict
        Values of pickable fields. The exact content is specific to the
        WorldObject triggering the event.
    clicks : int
        The total number of synchronous clicks.
    pointer_id : Any
        The value set by `set_pointer_capture()`.
    kwargs : Any
        Additional keyword arguments are forward to the :class:`base class
        <pygfx.objects.Event>`.

    Notes
    -----
    The values of this event follow the convention used by jupyter rfb. You can read
    about them `here <https://jupyter-rfb.readthedocs.io/en/stable/events.html>`_.

    """

    def __init__(
        self,
        *args,
        x,
        y,
        button=0,
        buttons=None,
        modifiers=None,
        ntouches=0,
        touches=None,
        pick_info=None,
        clicks=0,
        pointer_id=0,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.x = x
        self.y = y
        self.button = button
        self.buttons = buttons or ()
        self.modifiers = modifiers or ()
        self.ntouches = ntouches
        self.touches = touches or {}
        self.pick_info = pick_info or {}
        self.clicks = clicks
        self.pointer_id = pointer_id

    def copy(self, **kwargs):
        values = dict(
            type=self.type,
            x=self.x,
            y=self.y,
            button=self.button,
            buttons=self.buttons,
            modifiers=self.modifiers,
            ntouches=self.ntouches,
            touches=self.touches,
            pick_info=self.pick_info,
            clicks=self.clicks,
            pointer_id=self.pointer_id,
            bubbles=self.bubbles,
            target=self.target,
            cancelled=self.cancelled,
            time_stamp=self.time_stamp,
            root=self.root,
        )
        values.update(kwargs)
        return PointerEvent(**values)


class WheelEvent(PointerEvent):
    """Scrolling of the mouse wheel.

    Parameters
    ----------
    args : Any
        Positional arguments are forwarded to the :class:`base class
        <pygfx.objects.Event>`.
    dx : float
        The amount (in rad) by which the wheel was turned around the x-axis.
    dy : float
        The amount (in rad) by which the wheel was turned around the y-axis.
    kwargs : Any
        Additional keyword arguments are forward to the :class:`base class
        <pygfx.objects.Event>`.

    """

    def __init__(self, *args, dx, dy, **kwargs):
        super().__init__(*args, **kwargs)
        self.dx = dx
        self.dy = dy


class WindowEvent(Event):
    """Window resize event

    Parameters
    ----------
    args : Any
        Positional arguments are forwarded to the :class:`base class
        <pygfx.objects.Event>`.
    width : int
        The new width of the application window in screen space (px).
    height : int
        The new height of the application window in screen space (px).
    pixel_ratio : float
        The new ratio between logical pixels and physical pixels.
    kwargs : Any
        Additional keyword arguments are forward to the :class:`base class
        <pygfx.objects.Event>`.

    """

    def __init__(self, *args, width=None, height=None, pixel_ratio=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.width = width
        self.height = height
        self.pixel_ratio = pixel_ratio


class EventTarget:
    """Targetable object mixin.

    Mixin class that enables event handlers to be attached to objects
    of the mixed-in class.

    Parameters
    ----------
    args : Any
        Arguments are forwarded to allow multiple inheritance.
    kwargs : Any
        Kwargs are forwarded to allow multiple inheritance.

    """

    pointer_captures = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._event_handlers = defaultdict(set)

    def add_event_handler(self, *args):
        """Register an event handler.

        Arguments:
            callback (callable): The event handler. Must accept a
                single event argument.
            *types (list of strings): A list of event types.

        For the available event types, see
        https://jupyter-rfb.readthedocs.io/en/stable/events.html

        Can also be used as a decorator.

        Example:

        .. code-block:: py

            def my_handler(event):
                print(event)

            obj.add_event_handler(my_handler, "pointer_up", "pointer_down")

        Decorator usage example:

        .. code-block:: py

            @obj.add_event_handler("pointer_up", "pointer_down")
            def my_handler(event):
                print(event)
        """

        decorating = not callable(args[0])
        callback = None if decorating else args[0]
        types = args if decorating else args[1:]

        if not types:
            raise ValueError("No types registered for callback")
        if not all(isinstance(t, str) for t in types):
            raise TypeError("All types must be string.")

        def decorator(_callback):
            for type in types:
                self._event_handlers[type].add(_callback)
            return _callback

        if decorating:
            return decorator
        return decorator(callback)

    def remove_event_handler(self, callback, *types):
        """Unregister an event handler.

        Arguments:
            callback (callable): The event handler.
            *types (list of strings): A list of event types.
        """
        for type in types:
            self._event_handlers[type].remove(callback)

    def handle_event(self, event: Event):
        """Handle an incoming event.

        Arguments:
            event: The event to handle
        """
        event_type = event.type
        for callback in self._event_handlers[event_type].copy():
            if event.cancelled:
                break
            with log_exception(f"Error during handling {event_type} event"):
                callback(event)

    def set_pointer_capture(self, pointer_id, event_root):
        """Register this object to capture any other pointer events,
        until ``release_pointer_capture`` is called or an ``pointer_up``
        event is encountered.

        Arguments:
            pointer_id: id of pointer to capture (mouse, touch, etc.)
            event_root: the event root that this pointer is captured on
        """
        EventTarget.pointer_captures[pointer_id] = (
            ref(self),
            (event_root and ref(event_root)) or None,
        )

    def release_pointer_capture(self, pointer_id):
        """Release the pointer capture for the object that was registered
        to the given pointer_id.

        Arguments:
            pointer_id: id of pointer to release (mouse, touch, etc.)
        """
        EventTarget.pointer_captures.pop(pointer_id, None)


class RootEventHandler(EventTarget):
    """Pygfx event handler.

    Root event handler for the Pygfx event system.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Dictionary to track clicks, keyed on pointer_id
        self._click_tracker = {}
        # Dictionary to track targets, keyed on pointer_id
        self._target_tracker = {}

    def dispatch_event(self, event: Event):
        """Dispatch the given event.

        This method will dispatch an event by looking for the right target to
        handle the event. When a target is set on the event, then that target
        will be the first object that gets to handle the event. From there it
        will ask its parents one-by-one to handle the event as long as the event
        bubbles / propagates up or is not cancelled.

        The RootEventHandler object will serve as a virtual root for the tree
        hierarchy.

        Whenever an object has captured the pointer (for a specific pointer_id)
        then that object will get all pointer related events until the object
        releases the capture or a ``pointer_up`` event is encountered.

        This method will also keep track of ``pointer_down`` and ``pointer_up``
        events in order to generate and dispatch ``click`` and ``double_click``
        events.

        Parameters
        ----------
        event : Event
            The event to dispatch.

        """
        pointer_id = getattr(event, "pointer_id", None)

        # Check for captured pointer events
        if pointer_id is not None and pointer_id in EventTarget.pointer_captures:
            captured_target_ref, event_root_ref = EventTarget.pointer_captures[
                pointer_id
            ]
            captured_target, event_root = (
                captured_target_ref(),
                event_root_ref and event_root_ref(),
            )
            # If the pointer was captured in the context of another root event
            # handler, then let's not handle this event. It will be handled by
            # the appropriate RootEventHandler
            if event_root and event_root is not self:
                return

            if captured_target:
                # Encountered an event with pointer_id while there is a
                # capture active, so don't bubble, and retarget the event
                # to the captured target
                event._retarget(captured_target)
                event.stop_propagation()

        # Current target is either something that was under the pointer, or nothing
        # in which case we set the target to the root event handler (self)
        target = event.target or self

        # Update the target tracker on all `pointer_move` events
        if event.type == EventType.POINTER_MOVE:
            # Get the previous target for this pointer (if any)
            previous_target_ref = self._target_tracker.get(pointer_id)
            previous_target = (previous_target_ref and previous_target_ref()) or None
            # Check if the target has changed since the previous move event
            if previous_target is not target:
                # Update the current target for this pointer
                self._target_tracker[pointer_id] = (target and ref(target)) or None
                if previous_target is not None:
                    # Dispatch a `pointer_leave` event for the previous target
                    ev = event.copy(type="pointer_leave", target=previous_target)
                    self.dispatch_event(ev)
                # Dispatch a `pointer_enter` event for the new target
                ev = event.copy(type="pointer_enter")
                self.dispatch_event(ev)

        # Dispatch the event to target and bubble up the hierarchy
        while target:
            # Update the current target
            event._update_current_target(target)
            target.handle_event(event)
            # During handling of the event, the target might capture the pointer events
            if pointer_id is not None and pointer_id in EventTarget.pointer_captures:
                event._retarget(target)
                event.stop_propagation()
                if event.type == EventType.POINTER_UP:
                    captured_target.release_pointer_capture(pointer_id)
            if not event.bubbles or event.cancelled or target is self:
                break
            target = target.parent or self

        # Update the click tracker on all `pointer_down` events
        if event.type == EventType.POINTER_DOWN:
            tracked_click = self._click_tracker.get(pointer_id)
            # Check if the `pointer_id` is already tracked, targets
            # the same target and is within the DEBOUNCE time.
            # Bump the count and update the time_stamp if that is the case.
            # Otherwise, restart counting.
            if (
                tracked_click
                and (
                    (
                        tracked_click["target"] is not None
                        and tracked_click["target"]() is not None
                        and tracked_click["target"]() is event.target
                    )
                    or (tracked_click["target"] is None and event.target is None)
                )
                and event.time_stamp - tracked_click["time_stamp"] < CLICK_DEBOUNCE
            ):
                tracked_click["count"] += 1
                tracked_click["time_stamp"] = event.time_stamp
            else:
                self._click_tracker[pointer_id] = {
                    "count": 1,
                    "time_stamp": event.time_stamp,
                    "target": (event.target and ref(event.target)) or None,
                }
        # On all ``pointer_up`` events, see if the event is on the same target
        # as for the ``pointer_down``. If so, then a ``click`` event is dispatched.
        # When the counter for the click is at 2, then a ``double_click`` event
        # is dispatched.
        elif event.type == EventType.POINTER_UP:
            tracked_click = self._click_tracker.get(pointer_id)
            if tracked_click and (
                (
                    tracked_click["target"] is not None
                    and tracked_click["target"]() is not None
                    and tracked_click["target"]() is event.target
                )
                or (tracked_click["target"] is None and event.target is None)
            ):
                ev = event.copy(type="click", clicks=tracked_click["count"])
                self.dispatch_event(ev)
                if tracked_click["count"] == 2:
                    double_ev = event.copy(
                        type="double_click", clicks=tracked_click["count"]
                    )
                    self.dispatch_event(double_ev)



import numpy as np

from ._base import Geometry


def tetrahedron_geometry(radius=1.0, subdivisions=0):
    """Generate a tetrahedron.

    Creates a tetrahedron that is centered around local origin. Its vertices
    lie on the surface of a sphere of the given radius. Its faces are optionally
    subdivided if subdivisions>0.

    Parameters
    ----------
    radius : int
        The radius of a sphere that has the vertices on its surface.
    subdivisions: int
        The amount of times each face will be subdivided, where 0
        (the default) means no subdivision.

    Returns
    -------
    tetrahedron : Geometry
        A geometry object representing the desired tetrahedron.
        Mathematically, it consists of a set of open orientable manifolds.

    """
    positions = np.array(
        [
            [1, 1, 1],
            [-1, -1, 1],
            [-1, 1, -1],
            [1, -1, -1],
        ],
        dtype=np.float32,
    )

    indices = np.array(
        [
            [2, 0, 1],
            [0, 2, 3],
            [1, 0, 3],
            [2, 1, 3],
        ],
        dtype=np.int32,
    )

    return polyhedron_geometry(positions, indices, radius, subdivisions)


def octahedron_geometry(radius=1.0, subdivisions=0):
    """Generate a octahedron.

    Creates an octahedron that is centered around the local origin. It has its
    vertices lie on the surface of a sphere of the given radius. Its faces are
    optionally subdivided if subdivisions>0.

    Parameters
    ----------
    radius : int
        The radius of a sphere that has the vertices on its surface.
    subdivisions: int
        The amount of times each face will be subdivided, where 0
        (the default) means no subdivision.

    Returns
    -------
    octahedron : Geometry
        A geometry object representing the desired octahedron.
        Mathematically, it consists of a set of open orientable manifolds.

    """
    positions = np.array(
        [
            [1, 0, 0],
            [-1, 0, 0],
            [0, 1, 0],
            [0, -1, 0],
            [0, 0, 1],
            [0, 0, -1],
        ],
        dtype=np.float32,
    )

    indices = np.array(
        [
            [0, 4, 2],
            [0, 3, 4],
            [0, 5, 3],
            [0, 2, 5],
            [1, 5, 2],
            [1, 3, 5],
            [1, 4, 3],
            [1, 2, 4],
        ],
        dtype=np.int32,
    )

    return polyhedron_geometry(positions, indices, radius, subdivisions)


def icosahedron_geometry(radius=1.0, subdivisions=0):
    """Generate a icosahedron.

    Creates an icosahedron that is centered around the local origin. It has its
    vertices lie on the surface of a sphere of the given radius. Its faces are
    optionally subdivided if subdivisions>0.

    Parameters
    ----------
    radius : int
        The radius of a sphere that has the vertices on its surface.
    subdivisions: int
        The amount of times each face will be subdivided, where 0
        (the default) means no subdivision.

    Returns
    -------
    icosahedron : Geometry
        A geometry object representing the desired icosahedron.
        Mathematically, it consists of a set of open orientable manifolds.

    """
    t = (1 + np.sqrt(5)) / 2

    positions = np.array(
        [
            [-1, t, 0],
            [1, t, 0],
            [-1, -t, 0],
            [1, -t, 0],
            [0, -1, t],
            [0, 1, t],
            [0, -1, -t],
            [0, 1, -t],
            [t, 0, -1],
            [t, 0, 1],
            [-t, 0, -1],
            [-t, 0, 1],
        ],
        dtype=np.float32,
    )

    indices = np.array(
        [
            [0, 5, 11],
            [0, 1, 5],
            [0, 7, 1],
            [0, 10, 7],
            [0, 11, 10],
            [1, 9, 5],
            [5, 4, 11],
            [11, 2, 10],
            [10, 6, 7],
            [7, 8, 1],
            [3, 4, 9],
            [3, 2, 4],
            [3, 6, 2],
            [3, 8, 6],
            [3, 9, 8],
            [4, 5, 9],
            [2, 11, 4],
            [6, 10, 2],
            [8, 7, 6],
            [9, 1, 8],
        ],
        dtype=np.int32,
    )

    return polyhedron_geometry(positions, indices, radius, subdivisions)


def dodecahedron_geometry(radius=1.0, subdivisions=0):
    """Generate a dodecahedron.

    Creates an dodecahedron that is centered around the local origin. It has its
    vertices lie on the surface of a sphere of the given radius. Its faces are
    optionally subdivided if subdivisions>0.

    Parameters
    ----------
    radius : int
        The radius of a sphere that has the vertices on its surface.
    subdivisions: int
        The amount of times each face will be subdivided, where 0
        (the default) means no subdivision.

    Returns
    -------
    dodecahedron : Geometry
        A geometry object representing the desired dodecahedron.
        Mathematically, it consists of a set of open orientable manifolds.

    """
    t = (1 + np.sqrt(5)) / 2
    r = 1 / t

    positions = np.array(
        [
            [-1, -1, -1],
            [-1, -1, 1],
            [-1, 1, -1],
            [-1, 1, 1],
            [1, -1, -1],
            [1, -1, 1],
            [1, 1, -1],
            [1, 1, 1],
            [0, -r, -t],
            [0, -r, t],
            [0, r, -t],
            [0, r, t],
            [-r, -t, 0],
            [-r, t, 0],
            [r, -t, 0],
            [r, t, 0],
            [-t, 0, -r],
            [t, 0, -r],
            [-t, 0, r],
            [t, 0, r],
        ],
        dtype=np.float32,
    )

    indices = np.array(
        [
            [3, 7, 11],
            [3, 15, 7],
            [3, 13, 15],
            [7, 17, 19],
            [7, 6, 17],
            [7, 15, 6],
            [17, 8, 4],
            [17, 10, 8],
            [17, 6, 10],
            [8, 16, 0],
            [8, 2, 16],
            [8, 10, 2],
            [0, 1, 12],
            [0, 18, 1],
            [0, 16, 18],
            [6, 2, 10],
            [6, 13, 2],
            [6, 15, 13],
            [2, 18, 16],
            [2, 3, 18],
            [2, 13, 3],
            [18, 9, 1],
            [18, 11, 9],
            [18, 3, 11],
            [4, 12, 14],
            [4, 0, 12],
            [4, 8, 0],
            [11, 5, 9],
            [11, 19, 5],
            [11, 7, 19],
            [19, 14, 5],
            [19, 4, 14],
            [19, 17, 4],
            [1, 14, 12],
            [1, 5, 14],
            [1, 9, 5],
        ],
        dtype=np.int32,
    )

    return polyhedron_geometry(positions, indices, radius, subdivisions)


def polyhedron_geometry(
    positions: np.ndarray, indices: np.ndarray, radius: float, subdivisions: int
):
    """Generate a Polyhedron.

    Procedurally generate geometry representing a polyhedron, centered around
    the origin. The vertices lie on the surface of a sphere of the given radius.
    The faces are optionally subdivided if subdivisions>0.

    Parameters
    ----------
    positions : ndarray
        The vertices used to initialize polyhedron generation.
    indices : ndarray
        The face index used to initialize polyhedron generation.
    radius : float
        The vertices lie on the surface of a sphere with this radius.
    subdivisions : int
        The amount of times the faces will be subdivided, where 0 (the
        default) means no subdivision.

    Returns
    -------
    polyhedron : Geometry
        A geometry object representing the requested polyhedron.
        Mathematically, it consists of a set of open orientable manifolds.
    """
    # subdivide faces
    faces = positions[indices]
    a = faces[..., 0, :]
    ab = faces[..., 1, :] - a  # face winding convention
    ac = faces[..., 2, :] - a

    # assume we're creating a grid for vectorization purposes
    # and throw out half at the end...
    n = subdivisions + 2
    # set up the grid to have origin in the top right
    # so we can use numpy's upper triangle indexing
    i = np.linspace(1.0, 0.0, n, dtype=np.float32)
    j = np.linspace(0.0, 1.0, n, dtype=np.float32)
    alpha, beta = np.meshgrid(i, j)
    # grid shape is (n_faces, 3, n, n)
    grid = a[..., None, None] + alpha * ab[..., None, None] + beta * ac[..., None, None]

    # zero out all the elements in the lower corner of the grid
    # in other words those positions which lie outside of the original face
    grid = np.triu(grid)

    # normalize the positions so they lie on the surface a sphere
    # grid shape is (n_faces, n, n, 3)
    grid = np.moveaxis(grid, 1, 3)
    # keeping nan's so we can filter by that later
    with np.errstate(invalid="ignore"):
        grid /= np.linalg.norm(grid, axis=-1)[..., None]

    # create an index array for a single face's plane of quads
    indices = np.arange(n**2, dtype=np.uint32).reshape((n, n))
    index = np.empty((n - 1, n - 1, 2, 3), dtype=np.uint32)
    index[:, :, 0, 0] = indices[np.arange(n - 1)[:, None], np.arange(n - 1)[None, :]]
    # the remainder of the indices for every panel are relative
    # note that they are constructed such that the diagonal aligns
    # with that of np.triu; otherwise we get nan vertices in all
    # faces that include the diagonal
    index[:, :, 0, 1] = index[:, :, 0, 0] + 1
    index[:, :, 0, 2] = index[:, :, 0, 0] + n + 1
    index[:, :, 1, 0] = index[:, :, 0, 0] + n + 1
    index[:, :, 1, 1] = index[:, :, 1, 0] - 1
    index[:, :, 1, 2] = index[:, :, 1, 0] - n - 1

    # now index the grid
    grid = grid.reshape(grid.shape[0], -1, grid.shape[-1])
    # we get (n_faces, n-1, n-1, 2, 3 vertices, xyz)
    faces = grid[:, index]
    # and we reshape to get our new list of triangles
    faces = faces.reshape(-1, 3, 3)
    # filter out any triangles that contain nan elements
    nan_index = np.any(np.isnan(faces), axis=(-1, -2))
    faces = faces[~nan_index]
    positions = faces.reshape(-1, 3)

    # now that we have our subdivided faces with duplicated vertices
    # we can compute normals
    a = faces[..., 0, :]
    ab = faces[..., 1, :] - a  # face winding convention
    ac = faces[..., 2, :] - a

    ab = np.ndarray.astype(ab, np.float32, copy=False)
    ac = np.ndarray.astype(ac, np.float32, copy=False)
    face_normals = np.cross(ab, ac)

    face_normals /= np.linalg.norm(face_normals, axis=-1)[..., None]
    normals = np.broadcast_to(face_normals[..., None, :], faces.shape)

    # and our texcoords via spherical coordinates
    # NOTE: if the initial faces straddle the boundary between 2pi back to 0
    # you will get a weird seam. that can be handled but we don't do so for now
    x, y, z = np.moveaxis(faces, -1, 0)
    uu = np.arctan2(y, x) / (np.pi * 2)
    vv = np.arccos(z) / np.pi
    texcoords = np.stack([uu, vv], axis=-1)

    # after all this, scale the points to the radius
    # we don't do this earlier so we can rely on the assumption that vectors
    # are normalized there
    faces *= radius

    # technically if meshmaterial didn't require an index buffer we could leave this out
    indices = np.arange(positions.shape[0], dtype=np.int32).reshape(-1, 3)

    return Geometry(
        indices=indices,
        positions=positions,
        normals=normals.reshape(-1, 3),
        texcoords=texcoords.reshape(-1, 2),
    )



"""
Utilities for scraping examples to create a gallery.
"""

import os
import re
import time
from pathlib import Path

import imageio.v3 as iio
from sphinx_gallery.scrapers import figure_rst
from sphinx_gallery.py_source_parser import extract_file_config
from rendercanvas import BaseRenderCanvas

from ..renderers import Renderer


gallery_comment_pattern = re.compile(r"^# *sphinx_gallery_pygfx_docs *\=(.+)$", re.M)


def find_examples_for_gallery(examples_dir):
    """Find examples to include in the gallery.

    Examples are collected based on the value of the ``sphinx_gallery_pygfx_docs``
    comment in the example files:

    * If the comment is not found, the example is excluded.
    * ``sphinx_gallery_pygfx_docs = 'hidden'`` - not present in the gallery.
    * ``sphinx_gallery_pygfx_docs = 'code'`` - present only as code (don't run the code when building docs).
    * ``sphinx_gallery_pygfx_docs = 'screenshot'`` - present in the gallery with a screenshot.
    * ``sphinx_gallery_pygfx_docs = 'animate 3s 15fps'`` - present in the gallery with an animation (duration and fps are optional).

    Returns a dict that can be merged with the sphinx_gallery_conf.

    The Sphinx gallery does not have very good support to configure on a
    per-file basis. This utility function helps collect a lists of files, and
    pour these into regexp's for Sphinx to consume.
    """

    examples_dir = Path(examples_dir).absolute()

    examples_to_hide = []
    examples_to_show = []  # This is Sphinx' default; not actually used.
    examples_to_run = []

    # Collect files
    for filename in examples_dir.glob("**/*.py"):
        fname = str(filename.relative_to(examples_dir))
        example_code = filename.read_text(encoding="UTF-8")
        config = get_example_config(example_code)
        if config is None:
            examples_to_hide.append(fname)  # ignore files not having the comment
        elif not isinstance(config, str):
            raise RuntimeError(
                f"In '{fname}' expected sphinx_gallery_pygfx_docs to be a string."
            )
        elif config == "hidden":
            examples_to_hide.append(fname)
        elif config == "code":
            examples_to_show.append(fname)
        elif config.startswith(("screenshot", "animate")):
            examples_to_run.append(fname)
        else:
            raise RuntimeError(
                f"In '{fname}' got unexpected value for sphinx_gallery_pygfx_docs: '{config}'."
                + " Expecting 'hidden', 'code', 'screenshot' or 'animate'."
            )

    def patternize(path):
        return (os.sep + path).replace("\\", "\\\\").replace(".", "\\.")

    # Convert to regexp, because that's what Sphinx needs
    hide_pattern = "(" + "|".join(patternize(x) for x in examples_to_hide) + ")"
    run_pattern = "(" + "|".join(patternize(x) for x in examples_to_run) + ")"

    # Return dict that can be merged with the sphinx_gallery_conf
    return {
        "examples_dirs": str(examples_dir),
        "ignore_pattern": hide_pattern,
        "filename_pattern": run_pattern,
    }


def pygfx_scraper(block, block_vars, gallery_conf, **kwargs):
    """Scrape pygfx images and animations.

    Parameters
    ----------
    block : tuple
        A tuple containing the (label, content, line_number) of the block.
    block_vars : dict
        Dict of block variables.
    gallery_conf : dict
        Contains the configuration of Sphinx-Gallery
    **kwargs : dict
        Additional keyword arguments to pass to
        :meth:`~matplotlib.figure.Figure.savefig`, e.g. ``format='svg'``.
        The ``format`` kwarg in particular is used to set the file extension
        of the output file (currently only 'png', 'jpg', and 'svg' are
        supported).

    Returns
    -------
    rst : str
        The ReSTructuredText that will be rendered to HTML containing
        the images. This is often produced by :func:`figure_rst`.
    """

    src_file = block_vars["src_file"]
    namespace = block_vars["example_globals"]
    path_generator = block_vars["image_path_iterator"]

    # Get gallery config for this block. In config we already select files for
    # inclusion, so if this runs we know that we need a screenshot at least.
    # However, in scripts with multiple blocks, it can still happen that we
    # don't find a match for a specific block. In that case we default to
    # screenshot mode.
    config = get_example_config(block[1])
    if not config:
        return ""

    # Get canvas to sample the screenshot from
    canvas = select_canvas(src_file, namespace)

    config_parts = config.split()

    if config_parts[0] == "screenshot":
        # Configure
        config = {
            "lossless": True,
        }

        # Store as webp, even when lossless its ~50% of png
        img_filename = Path(next(path_generator)).with_suffix(".webp")
        render_screenshot(canvas, namespace, img_filename, **config)

    elif config_parts[0] == "animate":
        # Configure
        config = {
            "duration": 3.0,  # total duration of the animation, in seconds
            "fps": 20,  # frames per second
            "loop": 0,  # how many loops to show: 0 means loop forever
            "lossless": False,
        }
        for part in config_parts[1:]:
            val, unit = split_val_unit(part)
            if unit == "s":
                config["duration"] = val
            elif unit == "fps":
                config["fps"] = val
            else:
                raise RuntimeError(
                    f"In '{src_file}' got unexpected animation option: '{part}'"
                )

        # Store as webp, which is *much* smaller and better looking than gif
        img_filename = Path(next(path_generator)).with_suffix(".webp")
        render_movie(canvas, namespace, img_filename, **config)

    else:
        return ""

    return figure_rst([img_filename], gallery_conf["src_dir"])


def get_example_config(example_code):
    file_conf = extract_file_config(example_code)
    return file_conf.get("pygfx_docs", None)


def select_canvas(fname, namespace):
    """Select canvas from the given namespace."""
    canvas = None
    # Try get directly from canvas or renderer
    for target_name in ["renderer", "canvas"]:
        target = namespace.get(target_name, None)
        if target is None:
            pass
        elif isinstance(target, Renderer):
            canvas = target.target
            break
        elif isinstance(target, BaseRenderCanvas):
            canvas = target
            break
        elif target.__class__.__name__.endswith("WgpuCanvas"):
            canvas = target
            break
    # Try getting from proxy objects
    if canvas is None:
        for target_name in ["display", "disp", "plot", "figure"]:
            target = namespace.get(target_name, None)
            if hasattr(target, "canvas") and isinstance(
                target.canvas, BaseRenderCanvas
            ):
                canvas = target.canvas
                break
    # Found?
    if canvas is None:
        ns_keys = set(n for n in namespace.keys() if not n.startswith("_"))
        raise ValueError(
            f"In '{fname}' could not find object to sample the screenshot from."
            + f" Need either 'disp', 'renderer' or 'canvas'. Got namespace {ns_keys}"
        )
    return canvas


def render_screenshot(canvas, namespace, filename, *, lossless):
    # Render frame
    frame = canvas.draw()
    # Write image
    iio.imwrite(
        filename,
        frame,
        lossless=lossless,
    )


def render_movie(canvas, namespace, filename, *, duration, fps, loop, lossless):
    # Render frames
    frames = []
    n_frames = int(duration * fps)
    t0 = time.perf_counter()
    i = 0
    for i in range(n_frames):
        # In below lambda, the i is not bound, but thats ok
        namespace["perf_counter"] = lambda: t0 + i / fps  # noqa: B023
        frames.append(canvas.draw())
    # Write the video
    iio.imwrite(
        filename,
        frames,
        duration=round((1 / fps) * 1000),
        loop=loop,
        lossless=lossless,
    )


def split_val_unit(s):
    for i in range(len(s)):
        if s[i] not in "-+0.123456789":
            break
    return float(s[:i]), s[i:]



"""
_version.py v1.2

Simple version string management, using a hard-coded version string
for simplicity and compatibility, while adding git info at runtime.
See https://github.com/pygfx/_version for more info.
This code is subject to The Unlicense (public domain).

Any updates to this file should be done in https://github.com/pygfx/_version

Usage in short:

* Add this file to the root of your library (next to the `__init__.py`).
* On a new release, you just update the __version__.
"""

import logging
import subprocess
from pathlib import Path


# This is the base version number, to be bumped before each release.
# The build system detects this definition when building a distribution.
__version__ = "0.14.0"

# Set this to your library name
project_name = "pygfx"


logger = logging.getLogger(project_name)

# Get whether this is a repo. If so, repo_dir is the path, otherwise None.
repo_dir = Path(__file__).parents[1]
repo_dir = repo_dir if repo_dir.joinpath(".git").is_dir() else None


def get_version():
    """Get the version string."""
    if repo_dir:
        return get_extended_version()
    else:
        return __version__


def get_extended_version():
    """Get an extended version string with information from git."""

    release, post, labels = get_version_info_from_git()

    # Sample first 3 parts of __version__
    base_release = ".".join(__version__.split(".")[:3])

    # Check release
    if not release:
        release = base_release
    elif release != base_release:
        logger.warning(
            f"{project_name} version from git ({release})"
            + f" and __version__ ({base_release}) don't match."
        )

    # Build the total version
    version = release
    if post and post != "0":
        version += f".post{post}"
        if labels:
            version += "+" + ".".join(labels)
    elif labels and labels[-1] == "dirty":
        version += "+" + ".".join(labels)

    return version


def get_version_info_from_git():
    """Get (release, post, labels) from Git.

    With `release` the version number from the latest tag, `post` the
    number of commits since that tag, and `labels` a tuple with the
    git-hash and optionally a dirty flag.
    """

    # Call out to Git
    command = [
        "git",
        "describe",
        "--long",
        "--always",
        "--tags",
        "--dirty",
        "--first-parent",
    ]
    try:
        p = subprocess.run(command, cwd=repo_dir, capture_output=True)
    except Exception as e:
        logger.warning(f"Could not get {project_name} version: {e}")
        p = None

    # Parse the result into parts
    if p is None:
        parts = (None, None, "unknown")
    else:
        output = p.stdout.decode(errors="ignore")
        if p.returncode:
            stderr = p.stderr.decode(errors="ignore")
            logger.warning(
                f"Could not get {project_name} version.\n\nstdout: "
                + output
                + "\n\nstderr: "
                + stderr
            )
            parts = (None, None, "unknown")
        else:
            parts = output.strip().lstrip("v").split("-")
            if len(parts) <= 2:
                # No tags (and thus no post). Only git hash and maybe 'dirty'.
                parts = (None, None, *parts)

    # Return unpacked parts
    release, post, *labels = parts
    return release, post, labels


def _to_tuple(v):
    """Convert __version__ to version_info tuple."""
    v = __version__.split("+")[0]  # remove hash
    return tuple(int(i) if i.isnumeric() else i for i in v.split("."))


# Apply the versioning
base_version = __version__
__version__ = get_version()
version_info = _to_tuple(__version__)


# The CLI part

CLI_USAGE = """
_version.py

help            - Show this message.
version         - Show the current version.
bump VERSION    - Bump the __version__ to the given VERSION.
update          - Self-update the _version.py module by downloading the
                  reference code and replacing version number and project name.
""".lstrip()

if __name__ == "__main__":
    import sys
    import urllib.request

    _, *args = sys.argv

    if not args:
        print(f"{project_name} v{__version__}")

    elif args[0] == "version":
        print(f"{project_name} v{__version__}")

    elif args[0] == "bump":
        if len(args) != 2:
            sys.exit("Expected a version number to bump to.")
        new_version = args[1].lstrip("v")  # allow '1.2.3' and 'v1.2.3'
        if not new_version.count(".") == 2:
            sys.exit("Expected two dots in new version string.")
        if not all(s.isnumeric() for s in new_version.split(".")):
            sys.exit("Expected only numbers in new version string.")
        with open(__file__, "rb") as f:
            text = ref_text = f.read().decode()
        text = text.replace(base_version, new_version, 1)
        with open(__file__, "wb") as f:
            f.write(text.encode())
        print(f"Bumped version from '{base_version}' to '{new_version}'.")

    elif args[0] == "update":
        u = "https://raw.githubusercontent.com/pygfx/_version/main/_version.py"
        with urllib.request.urlopen(u) as f:
            text = ref_text = f.read().decode()
        text = text.replace("0.0.0", base_version, 1)
        text = text.replace("PROJECT_NAME", project_name, 1)
        with open(__file__, "wb") as f:
            f.write(text.encode())
        print("Updated to the latest _version.py.")

    elif args[0].lstrip("-") in ["h", "help"]:
        print(CLI_USAGE)

    else:
        print(f"Unknown command for _version.py: {args[0]!r}")
        print("Use ``python _version.py help`` to see a list of options.")



"""
Utils for the wgpu renderer.
"""

import json
from collections import OrderedDict

import wgpu

from ....utils.renderfunctionregistry import RenderFunctionRegistry


registry = RenderFunctionRegistry()


def register_wgpu_render_function(wobject_cls, material_cls):
    """Decorator for WGPU rendering functions.

    Parameters
    ----------
    wobject_cls : WorldObject
        The world object that this function knows how to render.
    material_cls : Material
        The world object that this function knows how to render.

    """
    # Note: in practice we use this to decorate shader classes, but the point is
    # that it registers a callable that - when called - produces one or more
    # BaseShader instances.

    def _register_wgpu_renderer(f):
        registry.register(wobject_cls, material_cls, f)
        return f

    return _register_wgpu_renderer


def nchannels_from_format(format):
    """Return the number of channels from a vertex-buffer format.

    Channels as in elements per item. I.e. will be 1 and 4 for a grayscale and
    rgba color buffer, respectively.
    """
    return int(to_vertex_format(format).partition("x")[2] or "1")


def to_index_format(format):
    """Convert any pygfx-allowed buffer format to the wgpu.IndexFormat."""
    if format in wgpu.IndexFormat:
        return format

    # Get primitive type
    primitives = {
        "i1": "sint8",
        "u1": "uint8",
        "i2": "sint16",
        "u2": "uint16",
        "i4": "sint32",
        "u4": "uint32",
        "f2": "float16",
        "f4": "float32",
    }
    primitive = primitives[format[-2:]]

    # Ignore dimensions; consider flattened array (indices is commonly Nx3).
    # Also ignore the sign, providing a signed array is fine as long as it does not contain negative numbers.
    fmt = primitive.replace("s", "u")

    if fmt in wgpu.IndexFormat:
        return fmt
    else:
        raise ValueError(f"Unexpected index format '{format}'")


def to_vertex_format(format):
    """Convert any pygfx-allowed buffer format to the wgpu.VertexFormat."""
    if format in wgpu.VertexFormat:
        return format

    # Get primitive type
    primitives = {
        "i1": "sint8",
        "u1": "uint8",
        "i2": "sint16",
        "u2": "uint16",
        "i4": "sint32",
        "u4": "uint32",
        "f2": "float16",
        "f4": "float32",
    }
    primitive = primitives[format[-2:]]

    fmt = ""
    if len(format) == 2:
        fmt = primitive
    elif len(format) == 4 and format[1] == "x":  # e.g. 3xf4
        if format[0] == "1":
            fmt = primitive
        elif format[0] in "234":
            fmt = primitive + "x" + str(format[0])

    if fmt in wgpu.VertexFormat:
        return fmt
    else:
        raise ValueError(f"Unexpected vertex format '{format}'")


def to_texture_format(format):
    """Convert any pygfx-allowed texture format to the wgpu.TextureFormat."""
    if format in wgpu.TextureFormat:
        return format

    # Note how we use normalized types (float in the shader) where we can,
    # because these types can work with an interpolating sampler.
    primitives = {
        "i1": "8snorm",
        "u1": "8unorm",
        "i2": "16sint",
        "u2": "16uint",
        "i4": "32sint",
        "u4": "32uint",
        "f2": "16float",
        "f4": "32float",
    }
    primitive = primitives[format[-2:]]

    fmt = ""
    if len(format) == 2:
        fmt = "r" + primitive
    elif len(format) == 4 and format[1] == "x":  # e.g. 3xf4
        if format[0] == "1":
            fmt = "r" + primitive
        elif format[0] == "2":
            fmt = "rg" + primitive
        elif format[0] == "3":
            fmt = "rgb" + primitive
        elif format[0] == "4":
            fmt = "rgba" + primitive

    if fmt in wgpu.TextureFormat:
        return fmt
    elif fmt.replace("rgb", "rgba") in wgpu.TextureFormat:
        return fmt  # We support rgb by wrapping in rgba textures, see renderers.wgpu/engine/update.py
    else:
        raise ValueError(f"Unexpected texture format '{format}'")


def generate_uniform_struct(dtype_struct, structname):
    """Generate wgsl code from a uniform struct defined with a numpy dtype."""
    code = f"""
        struct {structname} {{
    """.rstrip()

    # Obtain names of fields that are arrays or mat3. This is encoded as an empty field with a
    # name that starts with '__meta_xx__'
    array_names = []
    mat3_names = []
    for fieldname in dtype_struct.names:
        if fieldname.startswith("__meta_"):
            if fieldname.startswith("__meta_array_names__"):
                array_names.extend(fieldname.replace("__", " ").split()[1:])
            elif fieldname.startswith("__meta_mat3_names__"):
                mat3_names.extend(fieldname.replace("__", " ").split()[1:])

    # Process fields
    for fieldname, (dtype, offset) in dtype_struct.fields.items():
        # Note that dtype.names is a list of str names, but the dtype.fields also
        # includes keys for the field titles. We don't use titles, but keep in mind.
        if fieldname.startswith("__"):
            continue
        # Resolve primitive type
        primitive_type = dtype.base.name
        primitive_type = primitive_type.replace("float", "f")
        primitive_type = primitive_type.replace("uint", "u")
        primitive_type = primitive_type.replace("int", "i")
        # Resolve actual type (only scalar, vec, mat)
        shape = dtype.shape
        # Detect array
        length = -1
        if fieldname in array_names:
            length = shape[0]
            shape = shape[1:]
        # Obtain base type
        if shape == () or shape == (1,):
            # A scalar
            wgsl_type = align_type = primitive_type
        elif len(shape) == 1:
            # A vector
            n = shape[0]
            if n < 2 or n > 4:
                raise TypeError(f"Type {dtype} looks like an unsupported vec{n}.")
            wgsl_type = align_type = f"vec{n}<{primitive_type}>"
        elif len(shape) == 2:
            # A matNxM is Matrix of N columns and M rows
            # In wgsl, matrices are column-major; each vector in the matrix is a column.
            n, m = shape[0], shape[1]
            m_wgsl = m
            if n < 2 or n > 4 or m < 2 or m > 4:
                raise TypeError(f"Type {dtype} looks like an unsupported mat{n}x{m}.")
            if fieldname in mat3_names:
                assert m == 4
                m_wgsl = 3
            align_type = f"vec{m}<{primitive_type}>"
            wgsl_type = f"mat{n}x{m_wgsl}<{primitive_type}>"
        else:
            raise TypeError(f"Unsupported type {dtype}")
        # If an array, wrap it
        if length == 0:
            wgsl_type = align_type = None  # zero-length; dont use
        elif length > 0:
            wgsl_type = f"array<{wgsl_type},{length}>"
        else:
            pass  # not an array

        # Check alignment (https://www.w3.org/TR/WGSL/#alignment-and-size)
        if not wgsl_type:
            continue
        elif align_type == primitive_type:
            alignment = 4
        elif align_type.startswith("vec"):
            c = int(align_type.split("<")[0][-1])
            alignment = 8 if c < 3 else 16
        else:
            raise TypeError(f"Cannot establish alignment of wgsl type: {wgsl_type}")
        if offset % alignment != 0:
            # If this happens, our array_from_shadertype() has failed.
            raise TypeError(
                f"Struct alignment error: {structname}.{fieldname} alignment must be {alignment}"
            )

        code += f"\n            {fieldname}: {wgsl_type},"

    code += "\n        };"

    return code


class JsonEncoderWithWgpuSupport(json.JSONEncoder):
    def default(self, ob):
        if isinstance(ob, wgpu.GPUObjectBase):
            return ob.__class__.__name__ + "@" + hex(id(ob))


jsonencoder = JsonEncoderWithWgpuSupport()


def hash_from_value(value):
    """Simple way to create a hash from a (possibly compound) object.
    Assumes JSON encodable objects and GPU objects.
    """
    # Encode the value to string using json. The JSON encoder is so fast that
    # its hard to come up with something that can serialze to str faster.
    s = jsonencoder.encode(value)

    # Return hash (an int). For debugging purposes it can be helpul to return s instead.
    return hash(s)


class GpuCaches:
    """A collection of gpu caches."""

    def get_stats(self):
        """Get a dict mapping cache names to item counts."""
        d = {}
        for name, ob in self.__dict__.items():
            if isinstance(ob, GpuCache):
                d[name] = ob.get_stats()
        return d

    def enable(self):
        """Enable all caches."""
        for ob in self.__dict__.values():
            if isinstance(ob, GpuCache):
                ob.enable()

    def disable(self):
        """Disable all caches."""
        for ob in self.__dict__.values():
            if isinstance(ob, GpuCache):
                ob.disable()


gpu_caches = GpuCaches()


class GpuCache:
    """A cache for GPU objects using LRU eviction strategy."""

    def __init__(self, name, max_size=128):
        assert isinstance(name, str)
        assert not hasattr(gpu_caches, name)
        setattr(gpu_caches, name, self)

        self._objects = OrderedDict()
        self._max_size = max_size
        self._enabled = True
        self.hits = 0
        self.misses = 0

    def get_stats(self):
        """Get the number of (alive) objects in the cache."""
        return len(self._objects), self.hits, self.misses

    def enable(self):
        """Enable this cache."""
        self._enabled = True

    def disable(self):
        """Disable this cache."""
        self._enabled = False
        # Clear the cache when disabled to free memory
        self._objects.clear()

    def get(self, key):
        """Get the cached object or None."""
        if self._enabled:
            try:
                ob = self._objects[key]
                # Move to end (most recently used)
                self._objects.move_to_end(key)
                self.hits += 1
            except KeyError:
                ob = None
                self.misses += 1
        else:
            ob = None
        return ob

    def set(self, key, ob):
        """Store the given object under the given key.
        Uses LRU eviction when cache is full.
        """
        if not self._enabled:
            return

        if key in self._objects:
            # Update existing entry and move to end
            self._objects[key] = ob
            self._objects.move_to_end(key)
        else:
            # Add new entry
            if len(self._objects) >= self._max_size:
                # Remove least recently used item
                self._objects.popitem(last=False)
            self._objects[key] = ob

    def clear(self):
        """Clear all cached objects."""
        self._objects.clear()

    def set_max_size(self, max_size):
        """Set the maximum cache size."""
        self._max_size = max_size
        # Evict excess items if needed
        while len(self._objects) > self._max_size:
            self._objects.popitem(last=False)


class GfxSampler:
    """Simple wrapper for a GPUSampler. Should be considered read-only."""

    def __init__(self, filter="nearest", address_mode="clamp"):
        self.filter = filter
        self.address_mode = address_mode
        self._wgpu_object = None


class GfxTextureView:
    """Simple wrapper for a GPUTextureView. Should be considered read-only."""

    def __init__(
        self, texture, *, view_dim=None, layer_range=None, mip_range=None, aspect=None
    ):
        format = texture.format

        # Check view_dim
        default_view_dim = f"{texture.dim}d"
        if view_dim is None:
            view_dim = default_view_dim
            if texture.dim == 2 and texture.size[2] > 1:
                view_dim = "2d-array"
        elif isinstance(view_dim, int):
            view_dim = f"{view_dim}d"

        # Check layer_range (is half-open range)
        default_layer_range = 0, texture.size[2]
        if layer_range is None:
            layer_range = default_layer_range
        else:
            assert isinstance(layer_range, tuple) and len(layer_range) == 2

        # Check mip_range (is half-open range)
        if mip_range is None:
            mip_range = None
        else:
            assert isinstance(mip_range, tuple) and len(mip_range) == 2

        # Check aspect
        default_aspect = wgpu.TextureAspect.all
        if aspect is None:
            aspect = default_aspect
        else:
            assert aspect in wgpu.TextureAspect

        # Is this a default view on the texture?
        self.is_default_view = (
            view_dim == default_view_dim
            and layer_range == default_layer_range
            and aspect == default_aspect
        )

        # Store attributes
        self.texture = texture
        self.format = format
        self.view_dim = view_dim
        self.layer_range = layer_range
        self._mip_range = mip_range
        self.aspect = aspect
        self._wgpu_object = None

    @property
    def mip_range(self):
        # Calculated, since _wgpu_mip_level_count is set later
        if self._mip_range is None:
            return 0, self.texture._wgpu_mip_level_count
        else:
            return self._mip_range



"""
Utilities to load scenes from files, using trimesh.
"""

from importlib.util import find_spec
from typing import List

import numpy as np

import pygfx as gfx


def load_meshes(path, remote_ok=False):
    """Load meshes from a file.

    Deprecated: use load_mesh() instead.

    Parameters
    ----------
    path : str
        The location where the mesh is stored.
    remote_ok : bool
        Whether to allow loading meshes from URLs. Default is False.

    Returns
    -------
    meshes : list
        A list of loaded meshes.

    """
    raise DeprecationWarning("The load_meshes() function is replaced with load_mesh().")


def load_mesh(path, remote_ok=False):
    """Load mesh(es) from a file.

    This function requires the trimesh library.

    Parameters
    ----------
    path : str
        The location where the mesh or scene is stored.
        Can be a local file path or a URL.
    remote_ok : bool
        Whether to allow loading meshes from URLs. Default is False.

    Returns
    -------
    meshes : list
        A list of pygfx.Meshes.

    See Also
    --------
    load_scene
        Loads the entire scene (lights, graph, etc.) from a file.

    """
    if not find_spec("trimesh"):
        raise ImportError(
            "The `trimesh` library is required to load meshes: pip install trimesh"
        )

    import trimesh

    # Trimesh's load() performs a similar check and refers
    # loading from URLs to load_remote()
    if "https://" in str(path) or "http://" in str(path):
        if not remote_ok:
            raise ValueError(
                "Loading meshes from URLs is disabled. "
                "Set remote_ok=True to allow loading from URLs."
            )
        if not find_spec("httpx"):
            raise ImportError(
                "The `httpx` library is required to load meshes from URLs: pip install httpx"
            )
        scene = trimesh.load_remote(path)
    else:
        scene = trimesh.load(path)

    return meshes_from_trimesh(scene, apply_transforms=True)


def load_scene(
    path,
    flatten=False,
    meshes=True,
    materials=True,
    volumes=True,
    lights="auto",
    camera="auto",
    remote_ok=False,
):
    """Load file into a scene.

    When reading file formats that can contain more than just meshes (e.g. `.glb`) this function will
    attempt to import the entire scene, including lights, cameras, volumes and materials. If the file
    format is mesh-only (e.g. `.stl`) or volume-only (e.g. `.binvox`) we will instead construct a
    minimal scene.

    This function requires the trimesh library.

    Parameters
    ----------
    path : str
        The filepath. Can be a local file or a URL.
    flatten : bool
        If True, will ignore any hierarchical structure in the scene graph and
        instead import as a flat list of objects. Default is False.
    meshes : bool
        Whether to load meshes. Default is True.
    materials : bool
        Whether to load materials. Default is True.
    volumes : bool
        Whether to load 3D image volumes. Default is True.
    lights :  "auto" | "file" | "none"
        Whether to import lights. "auto" (default) will always add lights to the scene
        even if the file does not contain any; "file" will import only lights defined in
        the file; "none" will not add any lights to the scene.
    camera :  "auto" | "file" | "none"
        Whether to import camera. "auto" (default) will always add a camera to the scene
        even if the file does not contain any; "file" will import only cameras defined in
        the file; "none" will not add any cameras to the scene.
    remote_ok : bool
        Whether to allow loading files from URLs. Default is False.

    Returns
    -------
    scene : pygfx.Scene
        The loaded scene.

    See Also
    --------
    load_mesh
        Returns the flat meshes contained in a file.

    """
    if not find_spec("trimesh"):
        raise ImportError(
            "The `trimesh` library is required to load scenes: pip install trimesh"
        )

    if lights not in ("auto", "file", "none"):
        raise ValueError(f"Invalid value for `lights`: {lights}")

    if camera not in ("auto", "file", "none"):
        raise ValueError(f"Invalid value for `camera`: {camera}")

    import trimesh

    # Trimesh's load() performs a similar check and refers
    # loading from URLs to load_remote()
    if "https://" in str(path) or "http://" in str(path):
        if not remote_ok:
            raise ValueError(
                "Loading scenes from URLs is disabled. "
                "Set remote_ok=True to allow loading from URLs."
            )

        if not find_spec("httpx"):
            raise ImportError(
                "The `httpx` library is required to load meshes from URLs: pip install httpx"
            )
        tm_scene = trimesh.load_remote(path)
    else:
        tm_scene = trimesh.load(path)

    return scene_from_trimesh(
        tm_scene,
        flatten=flatten,
        meshes=meshes,
        volumes=volumes,
        materials=materials,
        lights=lights,
        camera=camera,
    )


def meshes_from_trimesh(scene, materials=True, apply_transforms=True):
    """Converts a trimesh scene into a flat list of pygfx Mesh objects.

    Parameters
    ----------
    scene : trimesh.Scene
        The scene to convert.
    materials : bool
        Whether to import materials. If False, a standard material will be created. Default is True.
    apply_transforms : bool
        Whether to apply the scene graph transforms directly to the meshes. Default is True.

    Returns
    -------
    meshes : list
        A list of loaded meshes (volumes are ignored).

    """
    return objects_from_trimesh(
        scene,
        meshes=True,
        volumes=False,
        materials=materials,
        apply_transforms=apply_transforms,
    )


def objects_from_trimesh(
    scene, meshes=True, volumes=True, materials=True, apply_transforms=True
):
    """Converts a trimesh scene into a flat list of pygfx objects.

    Parameters
    ----------
    scene : trimesh.Scene
        The scene to convert.
    meshes : bool
        Whether to import meshes. Default is True.
    volumes : bool
        Whether to load 3D image volumes. Default is True.
    materials : bool
        Whether to import materials for meshes. If False, a standard material will be created. Default is True.
    apply_transforms : bool
        Whether to apply the scene graph transforms directly to the meshes. Default is True.

    Returns
    -------
    objects : list
        A list of meshes/volumes.

    """
    import trimesh

    # NOTE: this function also loads volumes. Initially it was foreseen that
    # trimesh only supported meshes, but since it also loads volumes, we support that too.

    objects = []

    # If this is a single object we can just convert and return it
    if isinstance(scene, trimesh.Trimesh):
        m = gfx.Mesh(
            gfx.geometry_from_trimesh(scene),
            gfx.material_from_trimesh(scene),
        )
        objects.append(m)
    elif isinstance(scene, trimesh.voxel.VoxelGrid):
        vol = _volume_from_voxelgrid(scene)
        objects.append(vol)
    elif isinstance(scene, trimesh.Scene):
        # Scene consists of geometries (which in trimesh contain both
        # the vertices/faces as well as the mateiral) and a scene graph.
        # Each node in the graph consists of a geometry and a transform
        # A geometry can be used by more than one node in the graph.
        # Here we will convert each node in the graph to a pygfx.Mesh
        # but in the future it might be better to use pygfx.InstancedMesh.

        # Extract the geometries and materials
        gfx_geometries, gfx_materials = geometries_and_materials_from_trimesh(
            scene, materials=materials
        )

        # Generate a visual for each node
        for node_name in scene.graph.nodes_geometry:
            transform, geometry_name = scene.graph[node_name]
            gfx_cls = gfx.Mesh
            if gfx_materials[geometry_name].__class__.__name__.startswith("Volume"):
                gfx_cls = gfx.Volume
            ob = gfx_cls(gfx_geometries[geometry_name], gfx_materials[geometry_name])

            if apply_transforms:
                ob.local.matrix = transform  # set the node's transform

            objects.append(ob)
    else:
        raise NotImplementedError(
            f"Unexpected trimesh data: {scene.__class__.__name__}"
        )

    # Filter on type
    if meshes and volumes:
        return objects
    elif meshes:
        return [ob for ob in objects if isinstance(ob, gfx.Mesh)]
    elif volumes:
        return [ob for ob in objects if isinstance(ob, gfx.Volume)]
    else:
        return []


def geometries_and_materials_from_trimesh(scene, materials=True, de_duplicate=True):
    """Extract geometries and materials from a trimesh scene.

    Parameters
    ----------
    scene : trimesh.Scene
        The scene to convert.
    materials : bool
        Whether to import materials. If False, a default material will be created.
    de_duplicate : bool
        Whether to deduplicate geometries and materials. This may mean that
        multiple objects in the scene will share the same geometry and/or material.

    Returns
    -------
    gfx_geometries : dict
        A dictionary of geometry objects. Keys are the names of the geometries.
    gfx_materials : dict
        A dictionary of material objects. Keys are the names of the geometries.

    """
    import trimesh

    trimesh_to_gfx_geometries = {}  # cache for geometries
    trimesh_to_gfx_materials = {}  # cache for materials
    gfx_geometries = {}
    gfx_materials = {}
    for name, tm_ob in scene.geometry.items():
        # Convert object
        if de_duplicate and tm_ob in trimesh_to_gfx_materials:
            gfx_geometries[name] = trimesh_to_gfx_geometries[tm_ob]
        else:
            if isinstance(tm_ob, trimesh.Trimesh):
                gfx_geo = gfx.geometry_from_trimesh(tm_ob)
                if hasattr(tm_ob.visual, "material") and materials:
                    if (
                        de_duplicate
                        and tm_ob.visual.material in trimesh_to_gfx_materials
                    ):
                        gfx_materials[name] = trimesh_to_gfx_materials[
                            tm_ob.visual.material
                        ]
                    else:
                        gfx_materials[name] = trimesh_to_gfx_materials[
                            tm_ob.visual.material
                        ] = gfx.material_from_trimesh(tm_ob.visual.material)
                else:  # use a default material
                    gfx_materials[name] = gfx.MeshStandardMaterial()
            elif isinstance(tm_ob, trimesh.voxel.VoxelGrid):
                vol = _volume_from_voxelgrid(tm_ob)
                gfx_geo = vol.geometry
                gfx_materials[name] = vol.material
            else:
                raise NotImplementedError(f"{tm_ob.__class__.__name__}")

            gfx_geometries[name] = trimesh_to_gfx_geometries[tm_ob] = gfx_geo

    return gfx_geometries, gfx_materials


def scene_from_trimesh(
    tm_scene,
    flatten=False,
    meshes=True,
    materials=True,
    volumes=True,
    lights="auto",
    camera="auto",
    background=True,
):
    """Convert a trimesh scene into a pygfx scene.

    Parameters
    ----------
    tm_scene : trimesh.Scene
        A trimesh Scene.
    flatten : bool
        If True, will ignore any hierarchical structure in the scene graph and
        instead import as a flat list of objects. Default is False.
    meshes : bool
        Whether to import meshes. Default is True.
    materials : bool
        Whether to import materials for meshes. Default is True.
    volumes : bool
        Whether to load 3D image volumes. Default is True.
    lights :  "auto" | "file" | "none"
        Whether to import lights. "auto" (default) will always add lights to the scene
        even if the file does not contain any; "file" will import only lights defined in
        the file; "none" will not add any lights to the scene.
    camera :  "auto" | "file" | "none"
        Whether to import camera. "auto" (default) will always add a camera to the scene
        even if the file does not contain any; "file" will import only cameras defined in
        the file; "none" will not add any cameras to the scene.
    background : bool
        Trimesh scenes typically have a white background. If True, we will add a white background
        to the pygfx scene. If False, no background will be added. Default is True.

    Returns
    -------
    scene : pygfx.Scene
        The scene converted to pygfx.

    """
    if not find_spec("trimesh"):
        raise ImportError(
            "The `trimesh` library is required to load scenes: pip install trimesh"
        )

    if lights not in ("auto", "file", "none"):
        raise ValueError(f"Invalid value for `lights`: {lights}")

    if camera not in ("auto", "file", "none"):
        raise ValueError(f"Invalid value for `camera`: {camera}")

    import trimesh

    # Basic scene setup
    gfx_scene = gfx.Scene()

    # Convert single meshes into a scene (this makes the code below much simpler)
    if isinstance(tm_scene, trimesh.Trimesh):
        tm_scene = trimesh.Scene(geometry=tm_scene)
    elif isinstance(tm_scene, trimesh.voxel.VoxelGrid):
        # VoxelGrids can not be in a trimesh.Scene directly (trimesh renders them
        # as collection of cube meshes), so we need to handle them separately
        if volumes:
            vol = _volume_from_voxelgrid(tm_scene)
            gfx_scene.add(vol)

        # Generate an empty scene so we can continue with the rest of the function
        # (i.e. potentially add lights and camera)
        tm_scene = trimesh.Scene()
    elif not isinstance(tm_scene, trimesh.Scene):
        raise ValueError(f"Unexpected trimesh data: {type(tm_scene)}")

    # Extract a few things from the scene graph
    # `edges` is a dict {(parent, child): {'geometry': 'name', 'matrix': np.array}}
    edges = tm_scene.graph.transforms.edge_data
    # `leafs` is the list of leaf node names
    leafs = [
        n
        for n in tm_scene.graph.transforms.nodes
        if not tm_scene.graph.transforms.children.get(n, [])
    ]

    # Take care of meshes
    # Note: we're traversing the scene graph only for meshes (i.e. not lights, cameras, etc.)
    # That's because as far as I can tell, trimesh's scene graph only applies to geometries
    # while lights and cameras are stored separate from 'world'.
    if meshes and len(tm_scene.graph.nodes):
        if not flatten:
            # Load the geometries and materials
            gfx_geometries, gfx_materials = geometries_and_materials_from_trimesh(
                tm_scene, materials=materials
            )

            def _build_graph_bfs(node_name, node_object):
                """Recursively parse scene graph into pygfx.Groups."""
                # Note: not sure if that will ever happen in practice but
                # in theory, this implementation could run into the recursion
                # depth limit (3000 on my machine). If that turns out to be
                # a problem we need to switch the implementation

                # Go over this node's children
                for child_name in tm_scene.graph.transforms.children.get(node_name, []):
                    # Is this a leaf node?
                    is_leaf = child_name in leafs

                    # Does it have a geometry?
                    geometry_name = edges[(node_name, child_name)].get("geometry", None)
                    # See if this child has a geometry
                    if geometry_name is not None:
                        # Create the geometry
                        gfx_cls = gfx.Mesh
                        if gfx_materials[geometry_name].__class__.__name__.startswith(
                            "Volume"
                        ):
                            gfx_cls = gfx.Volume
                        child_object = gfx_cls(
                            gfx_geometries[geometry_name],
                            gfx_materials[geometry_name],
                        )
                    else:
                        child_object = gfx.Group()

                    # Apply matrix
                    child_object.local.matrix = edges[(node_name, child_name)]["matrix"]

                    # Connect child to parent unless this is a terminal group
                    if not is_leaf or not isinstance(child_object, gfx.Group):
                        node_object.add(child_object)

                    # If this is not a leaf node, we need to recurse
                    if not is_leaf:
                        _build_graph_bfs(child_name, child_object)

            # Recursively build the scene graph
            world = tm_scene.graph.base_frame
            gfx_scene.local.matrix = tm_scene.graph[world][0]
            _build_graph_bfs(world, gfx_scene)  # start at the root node
        # Just add the flat meshes
        else:
            # If we're flattening the scene graph, we'll just
            # convert all geometries in the scene to meshes
            gfx_scene.add(
                *objects_from_trimesh(
                    tm_scene,
                    apply_transforms=True,
                    meshes=meshes,
                    volumes=volumes,
                    materials=materials,
                )
            )

    # Take care of camera
    if camera != "none":
        # Accessing the .camera attribute directly will create
        # a camera if it doesn't exist, hence we check the
        # `.has_camera` attribute.
        # Also note that we're currently only using the
        # camera's transform but not e.g. FOV or resolution:
        if tm_scene.has_camera or camera == "auto":
            gfx_camera = gfx.PerspectiveCamera()
            gfx_scene.add(gfx_camera)

            # If a camera exists in the scene, use its transform
            if tm_scene.has_camera:
                gfx_camera.local.matrix = tm_scene.camera_transform
            # If no camera, make sure the camera actually looks
            # at the objects in the scene (if any)
            elif gfx_scene.get_bounding_box() is not None:
                gfx_camera.show_object(gfx_scene)

    # Take care of lights
    if lights != "none":
        # Parse lights. Similar to the camera, trimesh will create
        # lights automatically if we access the `.lights` attribute
        # directly.
        if hasattr(tm_scene, "_lights"):
            gfx_lights = []
            for light in tm_scene.lights:
                if isinstance(light, trimesh.scene.lighting.PointLight):
                    gfx_lights.append(gfx.PointLight())
                    gfx_lights[-1].distance = light.radius
                elif isinstance(light, trimesh.scene.lighting.DirectionalLight):
                    gfx_lights.append(gfx.DirectionalLight())
                elif isinstance(light, trimesh.scene.lighting.SpotLight):
                    gfx_lights.append(gfx.SpotLight())
                    gfx_lights[-1].distance = light.radius
                    gfx_lights[-1].angle = light.outerConeAngle
                else:
                    # Skip unknown light types
                    continue

                # These properties are common to all light types
                gfx_lights[-1].color = light.color / 255
                gfx_lights[-1].intensity = light.intensity
                gfx_lights[-1].local.matrix = tm_scene.graph[light.name][0]

            gfx_scene.add(*gfx_lights)
        # If no lights (either because the scene did not contain lights or because
        # `lights=False`), make sure things are actually visible
        elif lights == "auto":
            # Add an ambient light
            gfx_scene.add(gfx.AmbientLight())

            # Get the scene bounding box
            bbox = gfx_scene.get_bounding_box()

            # Calculate extent (if no geometries, bbox will be None)
            if bbox is not None:
                extent = bbox[1] - bbox[0]

                # Padd the bounding box
                bbox[0, :] -= extent
                bbox[1, :] += extent

                # Add point lights at the corners of the (padded) bounding box
                # This is ~ what trimesh does
                # If we don't pad, things look odd if the mesh(es) are boxy or flat
                for x, y, z in bbox:
                    light = gfx.PointLight()
                    (light.local.x, light.local.y, light.local.z) = (x, y, z)
                    gfx_scene.add(light)

    # By default trimesh scenes have a white background
    # while gfx.show() uses a black background
    if background:
        gfx_scene.add(gfx.Background(None, gfx.BackgroundMaterial((1, 1, 1))))

    return gfx_scene


def _volume_from_voxelgrid(vxl, cmap=None, clim="data"):
    """Helper function to convert a trimesh VoxelGrid into a pygfx Volume.

    Parameters
    ----------
    vxl : trimesh.voxels.VoxelGrid
        The voxel grid to convert.
    cmap : pygfx.Texture, optional
        The colormap to use. If None, this will default to `pygfx.cm.viridis`.
    clim : "data" | "dtype" | tuple | None
        The contrast limits to scale the data values with. By default ("data"), will
        use the min and max values of the data. If "dtype", will use the theoretical
        limits of the data type. If None, will use [0-1].

    Returns
    -------
    vol : pygfx.Volume
        The volume object.

    """
    import trimesh

    if not isinstance(vxl, trimesh.voxel.VoxelGrid):
        raise ValueError(f"Unexpected trimesh data: {type(vxl)}")

    # Extract the matrix. Note that trimesh uses xyz coordinates while
    # pygfx expects zyx - hence we transpose the matrix
    grid = vxl.matrix.T

    # Convert non-native byte order to native; e.g. >u4 -> u4 = uint64
    if grid.dtype.byteorder in (">", "<"):
        grid = grid.astype(grid.dtype.str.replace(grid.dtype.byteorder, ""))
    # Convert boolean matrices to uint16; I tried uint4 but that renders as
    # uniform volume and uint8 looks fuzzy
    elif grid.dtype == bool:
        grid = grid.astype(np.uint16)

    # Initialize texture
    tex = gfx.Texture(grid, dim=3)

    if cmap is None:
        cmap = gfx.cm.cividis

    # Find the potential min/max value of the volume
    if isinstance(clim, str):
        if clim == "data":
            clim = (grid.min(), grid.max())
        elif clim == "dtype":
            clim = (np.iinfo(grid.dtype).max, np.iinfo(grid.dtype).max)

    # Initialize the volume
    vol = gfx.Volume(
        gfx.Geometry(grid=tex),
        gfx.VolumeMipMaterial(clim=clim, map=cmap),
    )

    # Apply transform
    vol.local.matrix = vxl.transform

    return vol


def meshes_from_open3d(mesh_model, materials: bool = True) -> List[gfx.Mesh]:
    """
    Converts an Open3D TriangleMeshModel into a list of gfx.Mesh objects.

    Parameters
    ----------
    mesh_model : open3d.visualization.rendering.TriangleMeshModel
        The Open3D TriangleMeshModel to be converted.
    materials : bool, optional
        Whether to include materials in the conversion process. Default is True.

    Returns
    -------
    results : list of gfx.Mesh
        A list of gfx.Mesh objects converted from the input mesh model.

    Raises
    ------
    ImportError
        If the open3d library is not installed.
    ValueError
        If the input is not an Open3D TriangleMeshModel.
    """
    if not find_spec("open3d"):
        raise ImportError(
            "The `open3d` library is required for this function: pip install open3d"
        )

    import open3d

    if not isinstance(mesh_model, open3d.visualization.rendering.TriangleMeshModel):
        raise ValueError(f"Unexpected open3d data: {type(mesh_model)}")

    results = []

    model_materials: List[open3d.visualization.rendering.MaterialRecord] = (
        mesh_model.materials
    )
    mesh_info: open3d.visualization.rendering.TriangleMeshModel.MeshInfo

    for mesh_info in mesh_model.meshes:
        mesh: open3d.geometry.TriangleMesh = mesh_info.mesh
        material_idx: int = mesh_info.material_idx

        gfx_geometry = gfx.geometries.geometry_from_open3d(mesh)

        if materials and 0 <= material_idx < len(model_materials):
            material = model_materials[material_idx]
            gfx_material = gfx.materials.material_from_open3d(material)
        else:
            gfx_material = gfx.MeshStandardMaterial()

        gfx_mesh = gfx.Mesh(gfx_geometry, gfx_material)
        results.append(gfx_mesh)

    return results



# ruff: noqa

from PyInstaller.utils.hooks import collect_data_files

hiddenimports = ["pygfx.data_files"]
datas = []
datas += collect_data_files("pygfx.data_files")
datas += collect_data_files("pygfx.renderers.wgpu.wgsl")
